import asyncio
import base64
import hashlib
import os
import time
from datetime import datetime
from pathlib import Path
from typing import Any, AsyncIterator, Callable, Dict, List, NoReturn, Optional, Set
from unittest import mock

import pytest
from aiohttp import web
from yarl import URL

import neuromation.api.blob_storage
from neuromation.api import (
    Action,
    BlobListing,
    BucketListing,
    Client,
    PrefixListing,
    StorageProgressComplete,
    StorageProgressStart,
    StorageProgressStep,
)
from neuromation.api.blob_storage import calc_md5
from tests import _TestServerFactory


_MakeClient = Callable[..., Client]


FOLDER = Path(__file__).parent
DATA_FOLDER = FOLDER / "data"


class BlobUrlRotes:

    LIST_BUCKETS = r"/blob/b/"
    PUT_BUCKET = r"/blob/b/{bucket}"
    DELETE_BUCKET = r"/blob/b/{bucket}"

    LIST_OBJECTS = r"/blob/o/{bucket}"
    HEAD_OBJECT = r"/blob/o/{bucket}/{path:.+}"
    GET_OBJECT = r"/blob/o/{bucket}/{path:.+}"
    PUT_OBJECT = r"/blob/o/{bucket}/{path:.+}"
    DELETE_OBJECT = r"/blob/o/{bucket}/{path:.+}"


# Bucket `foo` structure
# foo
# ├── empty/
# ├── folder1/
# │   ├── xxx.txt
# │   └── yyy.json
# ├── folder2/
# │   └── big_file
# ├── test.json
# ├── test1.txt
# └── test2.txt


_ContentsObj = Dict[str, Dict[str, Any]]


def dir_list(path: Path) -> List[Dict[str, Any]]:
    # List all files and folders in directory
    res = []
    for item in path.glob("**/*"):
        if item.is_dir():
            res.append({"path": str(item.relative_to(path)), "dir": True})
        else:
            body = item.read_bytes().replace(b"\n\r", b"\n")
            res.append(
                {
                    "path": item.relative_to(path).as_posix(),
                    "dir": False,
                    "size": len(body),
                    "body": body,  # Fix Windows
                }
            )
    return res


@pytest.fixture
def blob_storage_contents() -> _ContentsObj:
    mtime1 = datetime(2019, 1, 1).timestamp()
    mtime2 = datetime(2019, 1, 2).timestamp()

    contents: Dict[str, Dict[str, Any]] = {
        "empty/": {
            "key": "empty/",
            "size": 0,
            "last_modified": mtime1,
            "body": b"",
        },
        "folder1/xxx.txt": {
            "key": "folder1/xxx.txt",
            "size": 1,
            "last_modified": mtime1,
            "body": b"w",
        },
        "folder1/yyy.json": {
            "key": "folder1/yyy.json",
            "size": 2,
            "last_modified": mtime2,
            "body": b"bb",
        },
        "test.json": {
            "key": "test.json",
            "size": 213,
            "last_modified": mtime1,
            "body": b"w" * 213,
        },
        "test1.txt": {
            "key": "test1.txt",
            "size": 111,
            "last_modified": mtime1,
            "body": b"w" * 111,
        },
        "test2.txt": {
            "key": "test2.txt",
            "size": 222,
            "last_modified": mtime2,
            "body": b"w" * 222,
        },
    }
    return contents


@pytest.fixture
async def blob_storage_server(
    aiohttp_server: _TestServerFactory,
    blob_storage_contents: _ContentsObj,
) -> Any:
    """Minimal functional Blob Storage server implementation"""
    CONTENTS = blob_storage_contents

    def with_keys(d: Dict[str, Any], keys: List[str]) -> Dict[str, Any]:
        res = {}
        for k, v in d.items():
            if k in keys:
                res[k] = v
        return res

    LIST_KEYS = ["key", "size", "last_modified"]

    app = web.Application()
    # fill route table

    async def list_buckets(request: web.Request) -> web.Response:
        assert "b3" in request.headers
        return web.json_response(
            [{"name": "foo", "creation_date": datetime(2019, 1, 1).isoformat()}]
        )

    async def list_blobs(request: web.Request) -> web.Response:
        assert "b3" in request.headers
        assert request.match_info["bucket"] == "foo"

        res: Dict[str, Any]
        prefix = request.query.get("prefix", "")
        contents = [
            with_keys(item, LIST_KEYS)
            for item in CONTENTS.values()
            if item["key"].startswith(prefix)
        ]
        common_prefixes: Set[str] = set()
        if request.query["recursive"] == "false":
            new_contents = []
            for item in contents:
                pos = item["key"].find("/", len(prefix))
                if pos != -1:
                    common_prefixes.add(item["key"][: pos + 1])
                else:
                    new_contents.append(item)
            contents = new_contents

        return web.json_response(
            {
                "contents": sorted(contents, key=lambda x: x["key"]),
                "common_prefixes": [{"prefix": p} for p in sorted(common_prefixes)],
                "is_truncated": False,
            }
        )

    async def get_blob(request: web.Request) -> web.StreamResponse:
        assert "b3" in request.headers
        assert request.match_info["bucket"] == "foo"

        key = request.match_info["path"]
        if key not in CONTENTS:
            raise web.HTTPNotFound()
        blob = CONTENTS[key]

        rng = request.http_range
        content = blob["body"]
        resp = web.StreamResponse(status=200)
        etag = hashlib.md5(blob["body"]).hexdigest()
        resp.headers.update({"ETag": repr(etag)})
        resp.last_modified = int(blob["last_modified"])  # type: ignore
        start, stop, _ = rng.indices(len(content))
        if not (rng.start is rng.stop is None):
            if start >= stop:
                raise RuntimeError
            resp.set_status(web.HTTPPartialContent.status_code)
            resp.headers["Content-Range"] = f"bytes {start}-{stop-1}/{len(content)}"
        resp.content_length = stop - start
        resp.content_type = "application/octet-stream"
        await resp.prepare(request)
        if request.method != "HEAD":
            chunk_size = max(len(content) * 2 // 5, 200)
            if stop - start > chunk_size:
                await resp.write(content[start : start + chunk_size])
                raise RuntimeError
            else:
                await resp.write(content[start:stop])
                await resp.write_eof()
        return resp

    async def put_blob(request: web.Request) -> web.Response:
        assert "b3" in request.headers
        assert request.match_info["bucket"] == "foo"

        key = request.match_info["path"]
        body = await request.content.read()
        blob = {
            "key": key,
            "size": len(body),
            "last_modified": time.time(),
            "body": body,
        }
        CONTENTS[key] = blob
        etag = hashlib.md5(body).hexdigest()

        return web.Response(headers={"ETag": repr(etag)})

    async def delete_blob(request: web.Request) -> NoReturn:
        assert "b3" in request.headers
        assert request.match_info["bucket"] == "foo"

        key = request.match_info["path"]
        if key not in CONTENTS:
            raise web.HTTPNotFound()

        del CONTENTS[key]

        raise web.HTTPNoContent()

    app = web.Application()
    app.router.add_get(BlobUrlRotes.LIST_BUCKETS, list_buckets)
    app.router.add_get(BlobUrlRotes.LIST_OBJECTS, list_blobs)
    # HEAD will also use this
    app.router.add_get(BlobUrlRotes.GET_OBJECT, get_blob)
    app.router.add_put(BlobUrlRotes.PUT_OBJECT, put_blob)
    app.router.add_delete(BlobUrlRotes.DELETE_OBJECT, delete_blob)

    return await aiohttp_server(app)


async def test_blob_storage_list_buckets(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    mtime1 = datetime.now()
    mtime2 = datetime.now()
    JSON = [
        {"name": "foo", "creation_date": mtime1.isoformat()},
        {"name": "bar", "creation_date": mtime2.isoformat()},
    ]

    async def handler(request: web.Request) -> web.Response:
        assert "b3" in request.headers
        assert request.path == BlobUrlRotes.LIST_BUCKETS
        assert request.query == {}
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get(BlobUrlRotes.LIST_BUCKETS, handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        ret = await client.blob_storage.list_buckets()

    assert ret == [
        BucketListing(
            name="foo",
            creation_time=int(mtime1.timestamp()),
            permission=Action.READ,
        ),
        BucketListing(
            name="bar",
            creation_time=int(mtime2.timestamp()),
            permission=Action.READ,
        ),
    ]


async def test_blob_storage_create_bucket(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    name = "my_bucket"

    async def handler(request: web.Request) -> web.Response:
        assert "b3" in request.headers
        assert request.path == BlobUrlRotes.PUT_BUCKET.format(bucket=name)
        assert request.query == {}
        return web.json_response({"location": "ua-east-1"})

    app = web.Application()
    app.router.add_put(BlobUrlRotes.PUT_BUCKET, handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        await client.blob_storage.create_bucket(name)


async def test_blob_storage_delete_bucket(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    name = "my_bucket"

    async def handler(request: web.Request) -> web.Response:
        assert "b3" in request.headers
        assert request.path == BlobUrlRotes.DELETE_BUCKET.format(bucket=name)
        assert request.query == {}
        return web.Response(status=204)

    app = web.Application()
    app.router.add_delete(BlobUrlRotes.DELETE_BUCKET, handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        await client.blob_storage.delete_bucket(name)


async def test_blob_storage_list_blobs(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    bucket_name = "foo"
    continuation_token = "cool_token"
    mtime1 = datetime.now()
    mtime2 = datetime.now()
    PAGE1_JSON = {
        "contents": [
            {"key": "test.json", "size": 213, "last_modified": mtime1.timestamp()}
        ],
        "common_prefixes": [{"prefix": "empty/"}, {"prefix": "folder1/"}],
        "is_truncated": True,
        "continuation_token": continuation_token,
    }
    PAGE2_JSON = {
        "contents": [
            {"key": "test1.txt", "size": 111, "last_modified": mtime1.timestamp()},
            {"key": "test2.txt", "size": 222, "last_modified": mtime2.timestamp()},
        ],
        "common_prefixes": [{"prefix": "folder2/"}],
        "is_truncated": False,
        "continuation_token": None,
    }

    async def handler(request: web.Request) -> web.Response:
        assert "b3" in request.headers
        assert request.path == BlobUrlRotes.LIST_OBJECTS.format(bucket=bucket_name)

        if "continuation_token" not in request.query:
            assert request.query["recursive"] == "false"
            assert request.query["max_keys"] in ("3", "6")
            return web.json_response(PAGE1_JSON)
        else:
            assert request.query == {
                "recursive": "false",
                "max_keys": "3",
                "continuation_token": continuation_token,
            }
            return web.json_response(PAGE2_JSON)

    app = web.Application()
    app.router.add_get(BlobUrlRotes.LIST_OBJECTS, handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        ret = await client.blob_storage.list_blobs(bucket_name, max_keys=3)

    assert ret == (
        [
            BlobListing(
                key="test.json",
                size=213,
                modification_time=int(mtime1.timestamp()),
                bucket_name=bucket_name,
            ),
        ],
        [
            PrefixListing(prefix="empty/", bucket_name=bucket_name),
            PrefixListing(prefix="folder1/", bucket_name=bucket_name),
        ],
    )

    async with make_client(srv.make_url("/")) as client:
        ret = await client.blob_storage.list_blobs(bucket_name, max_keys=6)

    assert ret == (
        [
            BlobListing(
                key="test.json",
                size=213,
                modification_time=int(mtime1.timestamp()),
                bucket_name=bucket_name,
            ),
            BlobListing(
                key="test1.txt",
                size=111,
                modification_time=int(mtime1.timestamp()),
                bucket_name=bucket_name,
            ),
            BlobListing(
                key="test2.txt",
                size=222,
                modification_time=int(mtime2.timestamp()),
                bucket_name=bucket_name,
            ),
        ],
        [
            PrefixListing(prefix="empty/", bucket_name=bucket_name),
            PrefixListing(prefix="folder1/", bucket_name=bucket_name),
            PrefixListing(prefix="folder2/", bucket_name=bucket_name),
        ],
    )


async def test_blob_storage_list_blobs_recursive(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    bucket_name = "foo"
    mtime1 = datetime.now()
    mtime2 = datetime.now()
    PAGE1_JSON = {
        "contents": [
            {"key": "folder1/xxx.txt", "size": 1, "last_modified": mtime1.timestamp()},
            {"key": "folder1/yyy.json", "size": 2, "last_modified": mtime2.timestamp()},
            {
                "key": "folder2/big_file",
                "size": 120 * 1024 * 1024,
                "last_modified": mtime1.timestamp(),
            },
        ],
        "common_prefixes": [],
        "is_truncated": False,
    }

    async def handler(request: web.Request) -> web.Response:
        assert "b3" in request.headers
        assert request.path == BlobUrlRotes.LIST_OBJECTS.format(bucket=bucket_name)
        assert request.query == {
            "recursive": "true",
            "max_keys": "1000",
            "prefix": "folder",
        }
        return web.json_response(PAGE1_JSON)

    app = web.Application()
    app.router.add_get(BlobUrlRotes.LIST_OBJECTS, handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        ret = await client.blob_storage.list_blobs(
            bucket_name, recursive=True, prefix="folder"
        )

    assert ret == (
        [
            BlobListing(
                key="folder1/xxx.txt",
                size=1,
                modification_time=int(mtime1.timestamp()),
                bucket_name=bucket_name,
            ),
            BlobListing(
                key="folder1/yyy.json",
                size=2,
                modification_time=int(mtime2.timestamp()),
                bucket_name=bucket_name,
            ),
            BlobListing(
                key="folder2/big_file",
                size=120 * 1024 * 1024,
                modification_time=int(mtime1.timestamp()),
                bucket_name=bucket_name,
            ),
        ],
        [],
    )


async def test_blob_storage_head_blob(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    bucket_name = "foo"
    key = "text.txt"
    mtime1 = time.time()

    async def handler(request: web.Request) -> web.StreamResponse:
        assert "b3" in request.headers
        assert request.path == f"/blob/o/{bucket_name}/{key}"
        assert request.match_info == {"bucket": bucket_name, "path": key}
        resp = web.StreamResponse(status=200)
        resp.headers.update({"ETag": '"12312908asd"'})
        resp.last_modified = int(mtime1)  # type: ignore
        resp.content_length = 111
        resp.content_type = "plain/text"
        return resp

    app = web.Application()
    app.router.add_head(BlobUrlRotes.HEAD_OBJECT, handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        ret = await client.blob_storage.head_blob(bucket_name, key=key)

    assert ret == BlobListing(
        key=key,
        size=111,
        modification_time=int(mtime1),
        bucket_name=bucket_name,
    )


async def test_blob_storage_get_blob(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    bucket_name = "foo"
    key = "text.txt"
    mtime1 = time.time()
    body = b"W" * 1000

    async def handler(request: web.Request) -> web.StreamResponse:
        assert "b3" in request.headers
        assert request.path == f"/blob/o/{bucket_name}/{key}"
        assert request.match_info == {"bucket": bucket_name, "path": key}
        resp = web.StreamResponse(status=200)
        resp.headers.update({"ETag": '"12312908asd"'})
        resp.last_modified = int(mtime1)  # type: ignore
        resp.content_length = len(body)
        resp.content_type = "plain/text"
        await resp.prepare(request)
        await resp.write(body)
        return resp

    app = web.Application()
    app.router.add_get(BlobUrlRotes.GET_OBJECT, handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        async with client.blob_storage.get_blob(bucket_name, key=key) as ret:
            assert ret.stats == BlobListing(
                key=key,
                size=1000,
                modification_time=int(mtime1),
                bucket_name=bucket_name,
            )
            assert await ret.body_stream.read() == body


async def test_blob_storage_fetch_blob(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    bucket_name = "foo"
    key = "text.txt"
    mtime1 = time.time()
    body = b"W" * 10 * 1024 * 1024

    async def handler(request: web.Request) -> web.StreamResponse:
        assert "b3" in request.headers
        assert request.path == f"/blob/o/{bucket_name}/{key}"
        assert request.match_info == {"bucket": bucket_name, "path": key}
        resp = web.StreamResponse(status=200)
        resp.headers.update({"ETag": '"12312908asd"'})
        resp.last_modified = int(mtime1)  # type: ignore
        resp.content_length = len(body)
        resp.content_type = "plain/text"
        await resp.prepare(request)
        await resp.write(body)
        return resp

    app = web.Application()
    app.router.add_get(BlobUrlRotes.GET_OBJECT, handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        buf = b""
        async for data in client.blob_storage.fetch_blob(bucket_name, key=key):
            buf += data
        assert buf == body


async def test_blob_storage_fetch_blob_partial_read(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    bucket_name = "foo"
    key = "text.txt"
    mtime1 = time.time()
    body = b"ababahalamaha"

    async def handler(request: web.Request) -> web.StreamResponse:
        assert "b3" in request.headers
        assert request.path == f"/blob/o/{bucket_name}/{key}"
        assert request.match_info == {"bucket": bucket_name, "path": key}
        rng = request.http_range
        start, stop, _ = rng.indices(len(body))
        resp = web.StreamResponse(status=web.HTTPPartialContent.status_code)
        resp.headers["ETag"] = '"12312908asd"'
        resp.headers["Content-Range"] = f"bytes {start}-{stop-1}/{len(body)}"
        resp.last_modified = int(mtime1)  # type: ignore
        resp.content_length = stop - start
        resp.content_type = "plain/text"
        await resp.prepare(request)
        await resp.write(body[start:stop])
        return resp

    app = web.Application()
    app.router.add_get(BlobUrlRotes.GET_OBJECT, handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        buf = bytearray()
        async for chunk in client.blob_storage.fetch_blob(bucket_name, key, 5):
            buf.extend(chunk)
        assert buf == b"halamaha"

        buf = bytearray()
        async for chunk in client.blob_storage.fetch_blob(bucket_name, key, 5, 4):
            buf.extend(chunk)
        assert buf == b"hala"

        buf = bytearray()
        async for chunk in client.blob_storage.fetch_blob(bucket_name, key, 5, 20):
            buf.extend(chunk)
        assert buf == b"halamaha"

        with pytest.raises(ValueError):
            async for chunk in client.blob_storage.fetch_blob(bucket_name, key, 5, 0):
                pass


async def test_blob_storage_fetch_blob_unsupported_partial_read(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    bucket_name = "foo"
    key = "text.txt"
    mtime1 = time.time()
    body = b"ababahalamaha"

    async def handler(request: web.Request) -> web.StreamResponse:
        assert "b3" in request.headers
        assert request.path == f"/blob/o/{bucket_name}/{key}"
        assert request.match_info == {"bucket": bucket_name, "path": key}
        resp = web.StreamResponse(status=200)
        resp.headers.update({"ETag": '"12312908asd"'})
        resp.last_modified = int(mtime1)  # type: ignore
        resp.content_length = len(body)
        resp.content_type = "plain/text"
        await resp.prepare(request)
        await resp.write(body)
        return resp

    app = web.Application()
    app.router.add_get(BlobUrlRotes.GET_OBJECT, handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        buf = b""
        async for data in client.blob_storage.fetch_blob(bucket_name, key):
            buf += data
        assert buf == body

        with pytest.raises(RuntimeError):
            async for data in client.blob_storage.fetch_blob(bucket_name, key, 5):
                pass


async def test_blob_storage_put_blob(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    bucket_name = "foo"
    key = "text.txt"
    body = b"W" * 10 * 1024 * 1024
    md5 = base64.b64encode(hashlib.md5(body).digest()).decode()
    etag = repr(hashlib.md5(body).hexdigest())

    async def handler(request: web.Request) -> web.StreamResponse:
        assert "b3" in request.headers
        assert request.path == f"/blob/o/{bucket_name}/{key}"
        assert request.match_info == {"bucket": bucket_name, "path": key}
        assert request.headers["X-Content-Length"] == str(len(body))
        assert await request.content.read() == body
        return web.Response(headers={"ETag": etag})

    app = web.Application()
    app.router.add_put(BlobUrlRotes.PUT_OBJECT, handler)

    srv = await aiohttp_server(app)

    async def async_iter() -> AsyncIterator[bytes]:
        yield body

    async with make_client(srv.make_url("/")) as client:
        resp_etag = await client.blob_storage.put_blob(
            bucket_name=bucket_name,
            key=key,
            body=async_iter(),
            size=len(body),
            content_md5=md5,
        )
        assert resp_etag == etag


async def test_blob_storage_delete_blob(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    bucket_name = "foo"
    key = "text.txt"

    async def handler(request: web.Request) -> web.StreamResponse:
        assert "b3" in request.headers
        assert request.path == f"/blob/o/{bucket_name}/{key}"
        return web.Response(status=204)

    app = web.Application()
    app.router.add_delete(BlobUrlRotes.DELETE_OBJECT, handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        res = await client.blob_storage.delete_blob(
            bucket_name=bucket_name,
            key=key,
        )
        assert res is None


async def test_blob_storage_calc_md5(tmp_path: Path) -> None:
    txt_file = tmp_path / "test.txt"
    body = b"""
    This is the greatest day of my whole life!!!
    """
    body_md5 = base64.b64encode(hashlib.md5(body).digest()).decode("ascii")
    with txt_file.open("wb") as f:
        f.write(body)
    assert (await calc_md5(txt_file))[0] == body_md5


async def test_blob_storage_large_calc_md5(tmp_path: Path) -> None:
    txt_file = tmp_path / "test.txt"
    size_mb = 20
    body = b"W" * (1024 * 1024)
    md5 = hashlib.md5()
    for _ in range(size_mb):
        md5.update(body)
    body_md5 = base64.b64encode(md5.digest()).decode("ascii")

    # Will write 100MB file
    with txt_file.open("wb") as f:
        for _ in range(size_mb):
            f.write(body)
    assert (await calc_md5(txt_file))[0] == body_md5


# high level API


async def test_blob_storage_upload_file_does_not_exists(
    make_client: _MakeClient,
) -> None:
    async with make_client("https://example.com") as client:
        with pytest.raises(FileNotFoundError):
            await client.blob_storage.upload_file(
                URL("file:///not-exists-file"), URL("blob://host/path/to/file.txt")
            )


async def test_blob_storage_upload_dir_doesnt_exist(make_client: _MakeClient) -> None:
    async with make_client("https://example.com") as client:
        with pytest.raises(IsADirectoryError):
            await client.blob_storage.upload_file(
                URL(FOLDER.as_uri()), URL("blob://host/path/to")
            )


async def test_blob_storage_upload_not_a_file(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
) -> None:

    file_path = Path(os.devnull).absolute()
    progress = mock.Mock()

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_file(
            URL(file_path.as_uri()), URL("blob:foo/file.txt"), progress=progress
        )

    uploaded = blob_storage_contents["file.txt"]
    assert uploaded["body"] == b""

    src = URL(file_path.as_uri())
    dst = URL("blob://default/foo/file.txt")
    progress.start.assert_called_with(StorageProgressStart(src, dst, 0))
    progress.step.assert_not_called()
    progress.complete.assert_called_with(StorageProgressComplete(src, dst, 0))


async def test_blob_storage_upload_regular_file_new_file(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
) -> None:
    file_path = DATA_FOLDER / "file.txt"
    file_size = file_path.stat().st_size
    progress = mock.Mock()

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_file(
            URL(file_path.as_uri()), URL("blob:foo/file.txt"), progress=progress
        )

    expected = file_path.read_bytes()
    uploaded = blob_storage_contents["file.txt"]
    assert uploaded["body"] == expected

    src = URL(file_path.as_uri())
    dst = URL("blob://default/foo/file.txt")
    progress.start.assert_called_with(StorageProgressStart(src, dst, file_size))
    progress.step.assert_called_with(
        StorageProgressStep(src, dst, file_size, file_size)
    )
    progress.complete.assert_called_with(StorageProgressComplete(src, dst, file_size))


async def test_blob_storage_upload_large_file(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
) -> None:
    local_dir = tmp_path / "local"
    local_dir.mkdir(exist_ok=True)
    local_file = local_dir / "file.txt"

    with local_file.open("wb") as f:
        for _ in range(1024):
            f.write(b"yncuNRzU0xhKSqIh" * (4 * 1024))

    progress = mock.Mock()

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_file(
            URL(local_file.as_uri()),
            URL("blob:foo/folder2/big_file"),
            progress=progress,
        )

    expected = local_file.read_bytes()
    uploaded = blob_storage_contents["folder2/big_file"]
    assert uploaded["body"] == expected

    src = URL(local_file.as_uri())
    dst = URL("blob://default/foo/folder2/big_file")
    file_size = len(expected)
    progress.start.assert_called_with(StorageProgressStart(src, dst, file_size))
    progress.step.assert_called_with(
        StorageProgressStep(src, dst, file_size, file_size)
    )
    progress.complete.assert_called_with(StorageProgressComplete(src, dst, file_size))


async def test_blob_storage_upload_regular_file_to_existing_file(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
) -> None:
    file_path = DATA_FOLDER / "file.txt"

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_file(
            URL(file_path.as_uri()), URL("blob:foo/test1.txt")
        )

    expected = file_path.read_bytes()
    uploaded = blob_storage_contents["test1.txt"]
    assert uploaded["body"] == expected


async def test_blob_storage_upload_regular_file_to_existing_dir_with_slash(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
) -> None:
    file_path = DATA_FOLDER / "file.txt"

    async with make_client(blob_storage_server.make_url("/")) as client:
        with pytest.raises(IsADirectoryError):
            await client.blob_storage.upload_file(
                URL(file_path.as_uri()), URL("blob:foo/empty/")
            )


async def test_blob_storage_upload_regular_file_to_existing_dir_without_slash(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
) -> None:
    file_path = DATA_FOLDER / "file.txt"

    async with make_client(blob_storage_server.make_url("/")) as client:
        with pytest.raises(IsADirectoryError):
            await client.blob_storage.upload_file(
                URL(file_path.as_uri()), URL("blob:foo/empty")
            )


async def test_blob_storage_upload_regular_file_to_existing_non_dir(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
) -> None:
    file_path = DATA_FOLDER / "file.txt"

    async with make_client(blob_storage_server.make_url("/")) as client:
        with pytest.raises(NotADirectoryError):
            await client.blob_storage.upload_file(
                URL(file_path.as_uri()), URL("blob:foo/test1.txt/subfile.txt")
            )


async def test_blob_storage_upload_regular_file_to_not_existing(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
) -> None:
    # In blob storage it's perfectly fine to upload on non-exising path
    file_path = DATA_FOLDER / "file.txt"

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_file(
            URL(file_path.as_uri()), URL("blob:foo/absent-dir/absent-file.txt")
        )

    expected = file_path.read_bytes()
    uploaded = blob_storage_contents["absent-dir/absent-file.txt"]
    assert uploaded["body"] == expected


async def test_blob_storage_upload_recursive_src_doesnt_exist(
    make_client: _MakeClient,
) -> None:
    async with make_client("https://example.com") as client:
        with pytest.raises(FileNotFoundError):
            await client.blob_storage.upload_dir(
                URL("file:does_not_exist"), URL("blob://host/path/to")
            )


async def test_blob_storage_upload_recursive_src_is_a_file(
    make_client: _MakeClient,
) -> None:
    file_path = DATA_FOLDER / "file.txt"

    async with make_client("https://example.com") as client:
        with pytest.raises(NotADirectoryError):
            await client.blob_storage.upload_dir(
                URL(file_path.as_uri()), URL("blob://host/path/to")
            )


async def test_blob_storage_upload_recursive_target_is_a_file(
    blob_storage_server: Any,
    make_client: _MakeClient,
) -> None:

    async with make_client(blob_storage_server.make_url("/")) as client:
        with pytest.raises(NotADirectoryError):
            await client.blob_storage.upload_dir(
                URL(DATA_FOLDER.as_uri()), URL("blob:foo/test1.txt")
            )


async def test_blob_storage_upload_empty_dir(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
) -> None:
    src_dir = tmp_path / "empty"
    src_dir.mkdir()
    assert list(src_dir.iterdir()) == []

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_dir(
            URL(src_dir.as_uri()), URL("blob:foo/folder")
        )

    assert "folder" not in blob_storage_contents
    uploaded = blob_storage_contents["folder/"]
    assert uploaded == {
        "key": "folder/",
        "size": 0,
        "last_modified": mock.ANY,
        "body": b"",
    }


async def test_blob_storage_upload_recursive_ok(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
) -> None:

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_dir(
            URL(DATA_FOLDER.as_uri()) / "nested", URL("blob:foo/folder")
        )

    keys = [v for k, v in blob_storage_contents.items() if k.startswith("folder/")]
    body = (DATA_FOLDER / "nested" / "folder" / "file.txt").read_bytes()
    assert keys == [
        {"key": "folder/", "size": 0, "last_modified": mock.ANY, "body": b""},
        {"key": "folder/folder/", "size": 0, "last_modified": mock.ANY, "body": b""},
        {
            "key": "folder/folder/file.txt",
            "size": len(body),
            "last_modified": mock.ANY,
            "body": body,
        },
    ]


async def test_blob_storage_upload_recursive_slash_ending(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
) -> None:

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_dir(
            URL(DATA_FOLDER.as_uri()) / "nested", URL("blob:foo/folder/")
        )

    keys = [v for k, v in blob_storage_contents.items() if k.startswith("folder/")]
    body = (DATA_FOLDER / "nested" / "folder" / "file.txt").read_bytes()
    assert keys == [
        {"key": "folder/", "size": 0, "last_modified": mock.ANY, "body": b""},
        {"key": "folder/folder/", "size": 0, "last_modified": mock.ANY, "body": b""},
        {
            "key": "folder/folder/file.txt",
            "size": len(body),
            "last_modified": mock.ANY,
            "body": body,
        },
    ]


async def test_blob_storage_upload_to_bucket_root(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
) -> None:

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_dir(
            URL(DATA_FOLDER.as_uri()) / "nested", URL("blob:foo/")
        )

    keys = [v for k, v in blob_storage_contents.items() if k.startswith("folder/")]
    body = (DATA_FOLDER / "nested" / "folder" / "file.txt").read_bytes()
    assert keys == [
        {"key": "folder/", "size": 0, "last_modified": mock.ANY, "body": b""},
        {
            "key": "folder/file.txt",
            "size": len(body),
            "last_modified": mock.ANY,
            "body": body,
        },
    ]


async def test_blob_storage_download_regular_file_to_absent_file(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
) -> None:
    local_dir = tmp_path / "local"
    local_dir.mkdir()
    local_file = local_dir / "file.txt"
    progress = mock.Mock()

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_file(
            URL("blob:foo/test1.txt"), URL(local_file.as_uri()), progress=progress
        )

    expected = blob_storage_contents["test1.txt"]["body"]
    downloaded = local_file.read_bytes()
    assert downloaded == expected

    src = URL("blob://default/foo/test1.txt")
    dst = URL(local_file.as_uri())
    file_size = len(expected)
    progress.start.assert_called_with(StorageProgressStart(src, dst, file_size))
    progress.step.assert_called_with(
        StorageProgressStep(src, dst, file_size, file_size)
    )
    progress.complete.assert_called_with(StorageProgressComplete(src, dst, file_size))


async def test_blob_storage_download_large_file(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
) -> None:
    local_dir = tmp_path / "local"
    local_dir.mkdir()
    local_file = local_dir / "file.txt"
    progress = mock.Mock()

    # 16 * 4 = 64 MB of data
    large_payload = b"yncuNRzU0xhKSqIh" * (4 * 1024 * 1024)
    blob_storage_contents["folder2/big_file"] = {
        "key": "folder2/big_file",
        "size": len(large_payload),
        "last_modified": datetime(2019, 1, 3).timestamp(),
        "body": large_payload,
    }

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_file(
            URL("blob:foo/folder2/big_file"),
            URL(local_file.as_uri()),
            progress=progress,
        )

    downloaded = local_file.read_bytes()
    assert downloaded == large_payload

    src = URL("blob://default/foo/folder2/big_file")
    dst = URL(local_file.as_uri())
    file_size = len(large_payload)
    progress.start.assert_called_with(StorageProgressStart(src, dst, file_size))
    progress.step.assert_called_with(
        StorageProgressStep(src, dst, file_size, file_size)
    )
    progress.complete.assert_called_with(StorageProgressComplete(src, dst, file_size))


async def test_blob_storage_download_regular_file_to_existing_file(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
) -> None:
    local_dir = tmp_path / "local"
    local_dir.mkdir()
    local_file = local_dir / "file.txt"
    local_file.write_bytes(b"Previous data")

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_file(
            URL("blob:foo/test1.txt"), URL(local_file.as_uri())
        )

    expected = blob_storage_contents["test1.txt"]["body"]
    downloaded = local_file.read_bytes()
    assert downloaded == expected


async def test_blob_storage_download_regular_file_to_dir(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
) -> None:
    local_dir = tmp_path / "local"
    local_dir.mkdir()

    async with make_client(blob_storage_server.make_url("/")) as client:
        # On Windows it will be signaled as PermissionError instead...
        with pytest.raises((IsADirectoryError, PermissionError)):
            await client.blob_storage.download_file(
                URL("blob:foo/test1.txt"), URL(local_dir.as_uri())
            )


async def test_blob_storage_download_regular_file_to_dir_slash_ended(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
) -> None:
    local_dir = tmp_path / "local"
    local_dir.mkdir()

    async with make_client(blob_storage_server.make_url("/")) as client:
        # On Windows it will be signaled as PermissionError instead...
        with pytest.raises((IsADirectoryError, PermissionError)):
            await client.blob_storage.download_file(
                URL("blob:foo/test1.txt"), URL(local_dir.as_uri() + "/")
            )


async def test_blob_storage_download_regular_file_to_non_file(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
) -> None:

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_file(
            URL("blob:foo/test1.txt"), URL(Path(os.devnull).absolute().as_uri())
        )


async def test_blob_storage_download_empty_dir(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
) -> None:
    target_dir = tmp_path / "empty"
    assert not target_dir.exists()

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_dir(
            URL("blob:foo/empty"), URL(target_dir.as_uri())
        )

    assert list(target_dir.iterdir()) == []


async def test_blob_storage_download_dir(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
) -> None:
    local_dir = tmp_path / "local"
    local_dir.mkdir()

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_dir(URL("blob:foo"), URL(local_dir.as_uri()))

    files = sorted(dir_list(local_dir), key=lambda x: x["path"])
    assert files == [
        {"dir": True, "path": "empty"},
        {"dir": True, "path": "folder1"},
        {"body": b"w", "dir": False, "path": "folder1/xxx.txt", "size": 1},
        {"body": b"bb", "dir": False, "path": "folder1/yyy.json", "size": 2},
        {"body": b"w" * 213, "dir": False, "path": "test.json", "size": 213},
        {"body": b"w" * 111, "dir": False, "path": "test1.txt", "size": 111},
        {"body": b"w" * 222, "dir": False, "path": "test2.txt", "size": 222},
    ]


async def test_blob_storage_download_dir_with_slash(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
) -> None:
    local_dir = tmp_path / "local"
    local_dir.mkdir()

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_dir(
            URL("blob:foo"), URL(local_dir.as_uri() + "/")
        )

    files = sorted(dir_list(local_dir), key=lambda x: x["path"])
    assert files == [
        {"dir": True, "path": "empty"},
        {"dir": True, "path": "folder1"},
        {"body": b"w", "dir": False, "path": "folder1/xxx.txt", "size": 1},
        {"body": b"bb", "dir": False, "path": "folder1/yyy.json", "size": 2},
        {"body": b"w" * 213, "dir": False, "path": "test.json", "size": 213},
        {"body": b"w" * 111, "dir": False, "path": "test1.txt", "size": 111},
        {"body": b"w" * 222, "dir": False, "path": "test2.txt", "size": 222},
    ]


@pytest.mark.parametrize(
    "pattern,expected_keys",
    [
        ("folder1/*", ["folder1/xxx.txt", "folder1/yyy.json"]),
        ("folder?/*", ["folder1/xxx.txt", "folder1/yyy.json", "folder2/big_file"]),
        ("**/*.json", ["folder1/yyy.json", "test.json"]),
        ("*/*.txt", ["folder1/xxx.txt"]),
        ("test[1-9].*", ["test1.txt", "test2.txt"]),
        ("test[2-3].*", ["test2.txt"]),
        # Should not match `/` for deep paths
        ("*.txt", ["test1.txt", "test2.txt"]),
        ("folder*", []),
        # Only glob style recursive supported:
        #   **/file
        #   folder/**/file
        #   folder/**
        ("folder**", []),
    ],
)
async def test_blob_storage_glob_blobs(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    pattern: str,
    expected_keys: List[str],
) -> None:
    bucket_name = "foo"
    blob_storage_contents["folder2/big_file"] = {
        "key": "folder2/big_file",
        "size": 2 * 1024,
        "last_modified": datetime(2019, 1, 2).timestamp(),
        "body": b"bb" * 1024,
    }

    async with make_client(blob_storage_server.make_url("/")) as client:
        ret = [
            x.key
            async for x in client.blob_storage.glob_blobs(bucket_name, pattern=pattern)
        ]
        assert ret == expected_keys


@pytest.fixture
def zero_time_threshold(monkeypatch: Any) -> None:
    monkeypatch.setattr(neuromation.api.blob_storage, "TIME_THRESHOLD", 0.0)


@pytest.fixture
def small_block_size(monkeypatch: Any) -> None:
    monkeypatch.setattr(neuromation.api.blob_storage, "READ_SIZE", 300)


async def test_storage_upload_file_update(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
    zero_time_threshold: None,
    small_block_size: None,
) -> None:
    # storage_file = storage_path / "file.txt"
    local_file = tmp_path / "file.txt"
    src = URL(local_file.as_uri())
    dst = URL("blob:foo/file.txt")

    # No destination file
    assert blob_storage_contents.get("file.txt") is None
    local_file.write_bytes(b"old content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_file(src, dst, update=True)
    assert blob_storage_contents["file.txt"]["body"] == b"old content"

    # Source file is newer
    local_file.write_bytes(b"new content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_file(src, dst, update=True)
    assert blob_storage_contents["file.txt"]["body"] == b"new content"

    # Destination file is newer, same size
    await asyncio.sleep(5)
    add_blob(blob_storage_contents, "file.txt", b"old")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_file(src, dst, update=True)
    assert blob_storage_contents["file.txt"]["body"] == b"old"


async def test_storage_upload_dir_update(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
    zero_time_threshold: None,
) -> None:
    local_dir = tmp_path / "folder"
    local_file = local_dir / "nested" / "file.txt"
    local_file.parent.mkdir(parents=True)
    src = URL(local_dir.as_uri())
    dst = URL("blob:foo")

    # No destination file
    assert blob_storage_contents.get("nested/file.txt") is None
    # assert not storage_file.exists()
    local_file.write_bytes(b"old content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_dir(src, dst, update=True)
    assert blob_storage_contents["nested/file.txt"]["body"] == b"old content"

    # Source file is newer
    local_file.write_bytes(b"new content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_dir(src, dst, update=True)
    assert blob_storage_contents["nested/file.txt"]["body"] == b"new content"

    # Destination file is newer, same size
    await asyncio.sleep(5)
    add_blob(blob_storage_contents, "nested/file.txt", b"old")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_dir(src, dst, update=True)
    assert blob_storage_contents["nested/file.txt"]["body"] == b"old"


async def test_storage_download_file_update(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
    zero_time_threshold: None,
) -> None:
    local_file = tmp_path / "file.txt"
    src = URL("blob:foo/file.txt")
    dst = URL(local_file.as_uri())

    # No destination file
    assert not local_file.exists()
    add_blob(blob_storage_contents, "file.txt", b"old content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_file(src, dst, update=True)
    assert local_file.read_bytes() == b"old content"

    # Source file is newer
    add_blob(blob_storage_contents, "file.txt", b"new content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_file(src, dst, update=True)
    assert local_file.read_bytes() == b"new content"

    # Destination file is newer
    await asyncio.sleep(2)
    local_file.write_bytes(b"old")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_file(src, dst, update=True)
    assert local_file.read_bytes() == b"old"


async def test_storage_download_file_continue(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
    zero_time_threshold: None,
    small_block_size: None,
) -> None:
    local_file = tmp_path / "file.txt"
    src = URL("blob:foo/file.txt")
    dst = URL(local_file.as_uri())

    # No destination file
    assert not local_file.exists()
    add_blob(blob_storage_contents, "file.txt", b"content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_file(src, dst, continue_=True)
    assert local_file.read_bytes() == b"content"

    # Source file is newer
    add_blob(blob_storage_contents, "file.txt", b"new content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_file(src, dst, continue_=True)
    assert local_file.read_bytes() == b"new content"

    # Destination file is newer, same size
    await asyncio.sleep(2)
    local_file.write_bytes(b"old content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_file(src, dst, continue_=True)
    assert local_file.read_bytes() == b"old content"

    # Destination file is shorter
    local_file.write_bytes(b"old")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_file(src, dst, continue_=True)
    assert local_file.read_bytes() == b"old content"

    # Destination file is longer
    local_file.write_bytes(b"old long content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_file(src, dst, continue_=True)
    assert local_file.read_bytes() == b"new content"


async def test_storage_download_dir_update(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
    zero_time_threshold: None,
) -> None:
    local_dir = tmp_path / "folder"
    local_file = local_dir / "nested" / "file.txt"
    src = URL("blob:foo")
    dst = URL(local_dir.as_uri())

    # No destination file
    assert not local_file.exists()
    add_blob(blob_storage_contents, "nested/file.txt", b"old content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_dir(src, dst, update=True)
    assert local_file.read_bytes() == b"old content"

    # Source file is newer
    add_blob(blob_storage_contents, "nested/file.txt", b"new content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_dir(src, dst, update=True)
    assert local_file.read_bytes() == b"new content"

    # Destination file is newer
    await asyncio.sleep(2)
    local_file.write_bytes(b"old")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_dir(src, dst, update=True)
    assert local_file.read_bytes() == b"old"


async def test_storage_download_dir_continue(
    blob_storage_server: Any,
    make_client: _MakeClient,
    blob_storage_contents: _ContentsObj,
    tmp_path: Path,
    zero_time_threshold: None,
) -> None:
    local_dir = tmp_path / "folder"
    local_file = local_dir / "nested" / "file.txt"
    src = URL("blob:foo")
    dst = URL(local_dir.as_uri())

    # No destination file
    assert not local_file.exists()
    add_blob(blob_storage_contents, "nested/file.txt", b"content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_dir(src, dst, continue_=True)
    assert local_file.read_bytes() == b"content"

    # Source file is newer
    add_blob(blob_storage_contents, "nested/file.txt", b"new content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_dir(src, dst, continue_=True)
    assert local_file.read_bytes() == b"new content"

    # Destination file is newer, same size
    await asyncio.sleep(2)
    local_file.write_bytes(b"old content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_dir(src, dst, continue_=True)
    assert local_file.read_bytes() == b"old content"

    # Destination file is shorter
    local_file.write_bytes(b"old")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_dir(src, dst, continue_=True)
    assert local_file.read_bytes() == b"old content"

    # Destination file is longer
    local_file.write_bytes(b"old long content")
    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.download_dir(src, dst, continue_=True)
    assert local_file.read_bytes() == b"new content"


async def test_storage_upload_dir_with_ignore_file_names(
    blob_storage_server: Any,
    make_client: _MakeClient,
    tmp_path: Path,
    blob_storage_contents: _ContentsObj,
) -> None:
    local_dir = tmp_path / "folder"
    local_dir2 = local_dir / "nested"
    local_dir2.mkdir(parents=True)
    for name in "one", "two", "three":
        (local_dir / name).write_bytes(b"")
        (local_dir2 / name).write_bytes(b"")
    (local_dir / ".neuroignore").write_text("one")
    (local_dir2 / ".gitignore").write_text("two")

    async with make_client(blob_storage_server.make_url("/")) as client:
        await client.blob_storage.upload_dir(
            URL(local_dir.as_uri()),
            URL("blob:foo/folder"),
            ignore_file_names={".neuroignore", ".gitignore"},
        )

    keys = sorted(k for k in blob_storage_contents if k.startswith("folder/"))
    assert keys == [
        "folder/",
        "folder/.neuroignore",
        "folder/nested/",
        "folder/nested/.gitignore",
        "folder/nested/three",
        "folder/three",
        "folder/two",
    ]


def add_blob(
    contents: Dict[str, Any], key: str, body: bytes, mtime: Optional[float] = None
) -> None:
    contents[key] = {}
    contents[key]["key"] = key
    contents[key]["body"] = body
    contents[key]["size"] = len(body)
    if mtime is None:
        mtime = time.time()
    contents[key]["last_modified"] = mtime



import abc
import operator
from typing import Sequence

from rich import box
from rich.console import RenderableType, RenderGroup
from rich.table import Table
from rich.text import Text

from neuromation.api import Disk
from neuromation.cli import utils
from neuromation.cli.formatters.jobs import format_datetime
from neuromation.cli.formatters.utils import URIFormatter


class BaseDisksFormatter:
    @abc.abstractmethod
    def __call__(self, jobs: Sequence[Disk]) -> RenderableType:
        pass


class SimpleDisksFormatter(BaseDisksFormatter):
    def __call__(self, disks: Sequence[Disk]) -> RenderableType:
        return RenderGroup(*[Text(disk.id) for disk in disks])


class DisksFormatter(BaseDisksFormatter):
    def __init__(
        self,
        uri_formatter: URIFormatter,
        *,
        long_format: bool = False,
    ) -> None:
        self._uri_formatter = uri_formatter
        self._long_format = long_format

    def _disk_to_table_row(self, disk: Disk) -> Sequence[str]:
        storage_str = utils.format_size(disk.storage)
        line = [disk.id, storage_str, self._uri_formatter(disk.uri), disk.status.value]
        if self._long_format:
            line += [format_datetime(disk.created_at), format_datetime(disk.last_usage)]
        return line

    def __call__(self, disks: Sequence[Disk]) -> RenderableType:
        disks = sorted(disks, key=operator.attrgetter("id"))
        table = Table(box=box.SIMPLE_HEAVY)
        # make sure that the first columnt is fully expanded
        width = len("disk-06bed296-8b27-4aa8-9e2a-f3c47b41c807")
        table.add_column("Id", style="bold", width=width)
        table.add_column("Storage")
        table.add_column("Uri")
        table.add_column("Status")
        if self._long_format:
            table.add_column("Created at")
            table.add_column("Last used")
        for disk in disks:
            table.add_row(*self._disk_to_table_row(disk))
        return table


class DiskFormatter:
    def __init__(self, uri_formatter: URIFormatter) -> None:
        self._uri_formatter = uri_formatter

    def __call__(self, disk: Disk) -> RenderableType:
        table = Table(
            box=None,
            show_header=False,
            show_edge=False,
        )
        table.add_column()
        table.add_column(style="bold")
        table.add_row("Id", disk.id)
        table.add_row("Storage", utils.format_size(disk.storage))
        table.add_row("Uri", self._uri_formatter(disk.uri))
        table.add_row("Status", disk.status.value)
        table.add_row("Created at", format_datetime(disk.created_at))
        table.add_row("Last used", format_datetime(disk.last_usage))
        return table



from typing import Awaitable, Callable

from aiohttp.test_utils import (
    RawTestServer,
    TestClient as _TestClient,
    TestServer as _TestServer,
)
from aiohttp.web import Application, Request, StreamResponse


_TestServerFactory = Callable[[Application], Awaitable[_TestServer]]
_RawTestServerFactory = Callable[
    [Callable[[Request], Awaitable[StreamResponse]]], Awaitable[RawTestServer]
]
_TestClientFactory = Callable[[Application], Awaitable[_TestClient]]



import abc
import contextlib
import enum
import operator
import os
import sys
import time
from abc import ABC
from dataclasses import dataclass
from fnmatch import fnmatch
from time import monotonic
from types import TracebackType
from typing import Any, Dict, Iterator, List, Sequence, Type

from rich.ansi import AnsiDecoder
from rich.columns import Columns
from rich.console import RenderableType
from rich.progress import (
    BarColumn,
    DownloadColumn,
    GetTimeCallable,
    Progress,
    TaskID,
    TextColumn,
    TransferSpeedColumn,
)
from rich.style import Style
from rich.table import Table
from rich.text import Text
from yarl import URL

from neuromation.api import (
    AbstractDeleteProgress,
    AbstractRecursiveFileProgress,
    Action,
    FileStatus,
    FileStatusType,
    StorageProgressComplete,
    StorageProgressDelete,
    StorageProgressEnterDir,
    StorageProgressFail,
    StorageProgressLeaveDir,
    StorageProgressStart,
    StorageProgressStep,
)
from neuromation.api.url_utils import _extract_path
from neuromation.cli.root import Root
from neuromation.cli.utils import format_size


RECENT_TIME_DELTA = 365 * 24 * 60 * 60 / 2
TIME_FORMAT = "%Y-%m-%d %H:%M:%S"


def chunks(list: Sequence[Any], size: int) -> Sequence[Any]:
    result = []
    for i in range(0, len(list), size):
        result.append(list[i : i + size])
    return result


def transpose(columns: Sequence[Sequence[Any]]) -> Sequence[Sequence[Any]]:
    height = len(columns)
    width = len(columns[0])
    result: Sequence[List[Any]] = [[] for _ in range(width)]
    for i in range(width):
        for j in range(height):
            if i < len(columns[j]):
                result[i].append(columns[j][i])
    return result


class GnuIndicators(str, enum.Enum):
    LEFT = "lc"
    RIGHT = "rc"
    END = "ec"
    RESET = "rs"
    NORM = "no"
    FILE = "fi"
    DIR = "di"
    LINK = "ln"
    FIFO = "pi"
    SOCKET = "so"
    BLK = "bd"
    CHR = "cd"
    MISSING = "mi"
    ORPHAN = "or"
    EXEC = "ex"
    DOOR = "do"
    SETUID = "su"
    SETGID = "sg"
    STICKY = "st"
    OTHER_WRITABLE = "ow"
    STICKY_OTHER_WRITABLE = "tw"
    CAP = "ca"
    MULTI_HARD_LINK = "mh"
    CLR_TO_EOL = "cl"


class ParseState(enum.Enum):
    PS_START = enum.auto()
    PS_LEFT = enum.auto()
    PS_ESCAPED = enum.auto()
    PS_ESCAPED_END = enum.auto()
    PS_RIGHT = enum.auto()
    PS_OCTAL = enum.auto()
    PS_HEX = enum.auto()
    PS_CARRET = enum.auto()


class BasePainter(abc.ABC):
    @abc.abstractmethod
    def paint(self, label: str, type: FileStatusType) -> Text:  # pragma: no cover
        pass


class NonePainter(BasePainter):
    def paint(self, label: str, type: FileStatusType) -> Text:
        return Text(label)


class GnuPainter(BasePainter):
    def __init__(self, ls_colors: str):
        self._defaults()
        self._parse_ls_colors(ls_colors)

    def _defaults(self) -> None:
        self.color_indicator: Dict[GnuIndicators, str] = {
            GnuIndicators.LEFT: "\033[",
            GnuIndicators.RIGHT: "m",
            GnuIndicators.END: "",
            GnuIndicators.RESET: "0",
            GnuIndicators.NORM: "",
            GnuIndicators.FILE: "",
            GnuIndicators.DIR: "01;34",
            GnuIndicators.LINK: "01;36",
            GnuIndicators.FIFO: "33",
            GnuIndicators.SOCKET: "01;35",
            GnuIndicators.BLK: "01;33",
            GnuIndicators.CHR: "01;33",
            GnuIndicators.MISSING: "",
            GnuIndicators.ORPHAN: "",
            GnuIndicators.EXEC: "01;32",
            GnuIndicators.DOOR: "01;35",
            GnuIndicators.SETUID: "37;41",
            GnuIndicators.SETGID: "30;43",
            GnuIndicators.STICKY: "37;44",
            GnuIndicators.OTHER_WRITABLE: "34;42",
            GnuIndicators.STICKY_OTHER_WRITABLE: "30;42",
            GnuIndicators.CAP: "30;41",
            GnuIndicators.MULTI_HARD_LINK: "",
            GnuIndicators.CLR_TO_EOL: "\033[K",
        }
        self.color_ext_type: Dict[str, str] = {}

    def _parse_ls_colors(self, ls_colors: str) -> None:
        def process(left: str, right: str) -> None:
            try:
                self.color_indicator[GnuIndicators(left)] = right
            except ValueError:
                self.color_ext_type[left] = right

        pos = 0
        left = right = escaped = ""
        num = 0
        state = ParseState.PS_START
        stack: List[ParseState] = []
        while pos < len(ls_colors):
            char = ls_colors[pos]
            if state == ParseState.PS_START:
                if char == ":":  # ignore colon
                    pos += 1
                else:
                    left = ""
                    state = ParseState.PS_LEFT
            elif state == ParseState.PS_OCTAL:
                if char in ["0", "1", "2", "3", "4", "5", "6", "7"]:
                    num = num * 8 + ord(char) - ord("0")
                    if num > 7:
                        state = ParseState.PS_ESCAPED_END
                        escaped = chr(num)
                    pos += 1
                else:
                    state = ParseState.PS_ESCAPED_END
                    escaped = chr(num)
            elif state == ParseState.PS_HEX:
                if char in ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]:
                    num = num * 16 + ord(char) - ord("0")
                    if num > 15:
                        state = ParseState.PS_ESCAPED_END
                        escaped = chr(num)
                    pos += 1
                elif char.upper() in ["A", "B", "C", "D", "E", "F"]:
                    num = num * 16 + 10 + ord(char.upper()) - ord("A")
                    if num > 15:
                        state = ParseState.PS_ESCAPED_END
                        escaped = chr(num)
                    pos += 1
                else:
                    state = ParseState.PS_ESCAPED_END
                    escaped = chr(num)

            elif state == ParseState.PS_ESCAPED_END:
                stack.pop()
                state = stack.pop()
                if state == ParseState.PS_LEFT:
                    left += escaped
                else:
                    right += escaped
                escaped = ""
            elif state == ParseState.PS_ESCAPED:
                if char in ["0", "1", "2", "3", "4", "5", "6", "7"]:
                    stack.append(state)
                    state = ParseState.PS_OCTAL
                    num = 0
                elif char.upper() == "X":
                    stack.append(state)
                    state = ParseState.PS_HEX
                    num = 0
                    pos += 1

                elif char == "a":
                    escaped = "\a"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "b":
                    escaped = "\b"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "e":
                    escaped = chr(27)
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "f":
                    escaped = "\f"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "n":
                    escaped = "\n"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "r":
                    escaped = "\r"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "t":
                    escaped = "\t"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "v":
                    escaped = "\v"
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "?":
                    escaped = chr(127)
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == "_":
                    escaped = " "
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
                elif char == chr(0):  # pragma: no cover
                    raise OSError("Cannot parse coloring scheme")
                else:
                    escaped = char
                    stack.append(state)
                    state = ParseState.PS_ESCAPED_END
                    pos += 1
            elif state == ParseState.PS_CARRET:
                if "@" <= char <= "~":
                    escaped = chr(ord(char) & 0o37)
                elif char == "?":
                    escaped = chr(127)
                else:
                    raise OSError("Cannot parse coloring scheme")
                stack.append(state)
                state = ParseState.PS_ESCAPED_END
                pos += 1

            elif state == ParseState.PS_LEFT:
                if char == "\\":
                    stack.append(state)
                    state = ParseState.PS_ESCAPED
                    pos += 1
                    escaped = ""
                elif char == "=":
                    right = ""
                    state = ParseState.PS_RIGHT
                    pos += 1
                elif char == "^":
                    stack.append(state)
                    state = ParseState.PS_CARRET
                    pos += 1
                    escaped = ""
                else:
                    left += char
                    pos = pos + 1
            elif state == ParseState.PS_RIGHT:
                if char == "\\":
                    stack.append(state)
                    state = ParseState.PS_ESCAPED
                    pos += 1
                    escaped = ""
                elif char == ":":
                    if right:
                        process(left, right)
                    state = ParseState.PS_START
                    pos += 1
                elif char == "^":
                    stack.append(state)
                    state = ParseState.PS_CARRET
                    pos += 1
                    escaped = ""
                else:
                    right += char
                    pos += 1

        if state == ParseState.PS_CARRET:
            raise OSError("Cannot parse coloring scheme")
        if state in [ParseState.PS_HEX, ParseState.PS_OCTAL]:
            escaped = chr(num)
            state = stack.pop()
        if state == ParseState.PS_ESCAPED:
            stack.append(ParseState.PS_ESCAPED)
            state = ParseState.PS_ESCAPED_END
        if state == ParseState.PS_ESCAPED_END:
            stack.pop()
            state = stack.pop()
            if state == ParseState.PS_RIGHT:  # pragma no branch
                right += escaped
        if state == ParseState.PS_RIGHT and len(right):
            process(left, right)

    def paint(self, label: str, type: FileStatusType) -> Text:
        mapping = {
            FileStatusType.FILE: self.color_indicator[GnuIndicators.FILE],
            FileStatusType.DIRECTORY: self.color_indicator[GnuIndicators.DIR],
        }
        color = mapping[type]
        if not color:
            color = self.color_indicator[GnuIndicators.NORM]
        if type == FileStatusType.FILE:
            for pattern, value in self.color_ext_type.items():
                if fnmatch(label, pattern):
                    color = value
                    break
        if color:
            # If formatted text contains newlines for some reason,
            # painter should escape them
            return Text("\\n").join(
                AnsiDecoder().decode(
                    self.color_indicator[GnuIndicators.LEFT]
                    + color
                    + self.color_indicator[GnuIndicators.RIGHT]
                    + label
                    + self.color_indicator[GnuIndicators.LEFT]
                    + self.color_indicator[GnuIndicators.RESET]
                    + self.color_indicator[GnuIndicators.RIGHT]
                )
            )
        else:
            return Text(label)


class BSDAttributes(enum.Enum):
    DIRECTORY = 1
    LINK = 2
    SOCKET = 3
    PIPE = 4
    EXECUTABLE = 5
    BLOCK = 6
    CHARACTER = 7
    EXECUTABLE_SETUID = 8
    EXECUTABLE_SETGID = 9
    DIRECTORY_WRITABLE_OTHERS_WITH_STICKY = 10
    DIRECTORY_WRITABLE_OTHERS_WITHOUT_STICKY = 11


class BSDPainter(BasePainter):
    def __init__(self, lscolors: str):
        self._parse_lscolors(lscolors)

    def _parse_lscolors(self, lscolors: str) -> None:
        parts = chunks(lscolors, 2)
        self._colors: Dict[BSDAttributes, str] = {}
        num = 0
        for attr in BSDAttributes:
            self._colors[attr] = parts[num]
            num += 1

    def paint(self, label: str, type: FileStatusType) -> Text:
        color = ""
        if type == FileStatusType.DIRECTORY:
            color = self._colors[BSDAttributes.DIRECTORY]
        if color:
            char_to_color = {
                "a": "black",
                "b": "red",
                "c": "green",
                "d": "brown",
                "e": "blue",
                "f": "magenta",
                "g": "cyan",
                "h": "white",
            }
            bold = None
            fg = bg = None
            if color[0].lower() in char_to_color.keys():
                fg = char_to_color[color[0].lower()]
                if color[0].isupper():
                    bold = True
            if color[1] in char_to_color.keys():
                bg = char_to_color[color[1]]
            if fg or bg or bold:
                return Text(
                    label,
                    style=Style(color=fg, bgcolor=bg, bold=bold),
                )
        return Text(label)


def get_painter(color: bool) -> BasePainter:
    if color:
        ls_colors = os.getenv("LS_COLORS")
        if ls_colors:
            return GnuPainter(ls_colors)
        lscolors = os.getenv("LSCOLORS")
        if lscolors:
            return BSDPainter(lscolors)
    return NonePainter()


class BaseFilesFormatter:
    @abc.abstractmethod
    def __call__(
        self, files: Sequence[FileStatus]
    ) -> RenderableType:  # pragma: no cover
        pass


class LongFilesFormatter(BaseFilesFormatter):
    permissions_mapping = {Action.MANAGE: "m", Action.WRITE: "w", Action.READ: "r"}

    file_types_mapping = {FileStatusType.FILE: "-", FileStatusType.DIRECTORY: "d"}

    def __init__(self, human_readable: bool, color: bool):
        self.human_readable = human_readable
        self.painter = get_painter(color)

    def _columns_for_file(self, file: FileStatus) -> Sequence[RenderableType]:

        type = self.file_types_mapping[file.type]
        permission = self.permissions_mapping[file.permission]

        date = time.strftime(TIME_FORMAT, time.localtime(file.modification_time))

        if self.human_readable:
            size = format_size(file.size).rstrip("B")
        else:
            size = str(file.size)

        name = self.painter.paint(file.name, file.type)

        return [f"{type}{permission}", f"{size}", f"{date}", name]

    def __call__(self, files: Sequence[FileStatus]) -> RenderableType:
        table = Table.grid(padding=(0, 2))
        table.add_column()  # Type/Permissions
        table.add_column(justify="right")  # Size
        table.add_column()  # Date
        table.add_column()  # Filename

        for file in files:
            table.add_row(*self._columns_for_file(file))
        return table


class SimpleFilesFormatter(BaseFilesFormatter):
    def __init__(self, color: bool):
        self.painter = get_painter(color)

    def __call__(self, files: Sequence[FileStatus]) -> RenderableType:
        return Text("\n").join(
            self.painter.paint(file.name, file.type) for file in files
        )


class VerticalColumnsFilesFormatter(BaseFilesFormatter):
    def __init__(self, width: int, color: bool):
        self.width = width
        self.painter = get_painter(color)

    def __call__(self, files: Sequence[FileStatus]) -> RenderableType:
        if not files:
            return ""
        items = [self.painter.paint(file.name, file.type) for file in files]
        column_width = max(len(item) for item in items) + 1
        return Columns(
            [self.painter.paint(file.name, file.type) for file in files],
            align="left",
            column_first=True,
            width=column_width,
            padding=0,
        )


class FilesSorter(str, enum.Enum):
    NAME = "name"
    SIZE = "size"
    TIME = "time"

    def key(self) -> Any:
        field = None
        if self == self.NAME:
            field = "name"
        elif self == self.SIZE:
            field = "size"
        elif self == self.TIME:
            field = "modification_time"
        assert field
        return operator.attrgetter(field)


# progress indicator


class DeleteProgress(AbstractDeleteProgress):
    def __init__(self, root: Root) -> None:
        self._root = root
        self.painter = get_painter(root.color)

    def fmt_url(self, url: URL, type: FileStatusType) -> Text:
        return self.painter.paint(str(url), type)

    def delete(self, data: StorageProgressDelete) -> None:
        url_label = self.fmt_url(
            data.uri,
            FileStatusType.DIRECTORY if data.is_dir else FileStatusType.FILE,
        )
        self._root.print(f"Removed: {url_label}")


class StorageProgressContextManager:
    def __init__(self, progress: "BaseStorageProgress"):
        self._progress = progress

    def begin(self, src: URL, dst: URL) -> "BaseStorageProgress":
        self._progress.begin(src, dst)
        return self._progress

    def __enter__(self) -> "BaseStorageProgress":
        return self._progress

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        self._progress.end()


class BaseStorageProgress(AbstractRecursiveFileProgress, ABC):
    def begin(
        self, src: URL, dst: URL
    ) -> StorageProgressContextManager:  # pragma: no cover
        return StorageProgressContextManager(self)

    def end(self) -> None:  # pragma: no cover
        pass


def create_storage_progress(
    root: Root,
    show_progress: bool,
    *,
    get_time: GetTimeCallable = time.monotonic,
    auto_refresh: bool = True,  # only disabled in test
) -> BaseStorageProgress:
    if show_progress:
        return TTYProgress(root, get_time=get_time, auto_refresh=auto_refresh)
    else:
        return StreamProgress(root)


def format_url(url: URL) -> str:
    if url.scheme == "file":
        path = _extract_path(url)
        return str(path)
    else:
        return str(url)


class StreamProgress(BaseStorageProgress):
    def __init__(self, root: Root) -> None:
        self.painter = get_painter(root.color)
        self.verbose = root.verbosity > 0
        self._root = root

    def fmt_url(self, url: URL, type: FileStatusType) -> Text:
        label = format_url(url)
        return self.painter.paint(label, type)

    def begin(self, src: URL, dst: URL) -> StorageProgressContextManager:
        if self.verbose:
            src_label = self.fmt_url(src, FileStatusType.DIRECTORY)
            dst_label = self.fmt_url(dst, FileStatusType.DIRECTORY)
            self._root.print(f"Copy {src_label} -> {dst_label}")
        return super().begin(src, dst)

    def start(self, data: StorageProgressStart) -> None:
        pass

    def complete(self, data: StorageProgressComplete) -> None:
        if not self.verbose:
            return
        src = self.fmt_url(data.src, FileStatusType.FILE)
        dst = self.fmt_url(data.dst, FileStatusType.FILE)
        self._root.print(f"{src} -> {dst}")

    def step(self, data: StorageProgressStep) -> None:
        pass

    def enter(self, data: StorageProgressEnterDir) -> None:
        if not self.verbose:
            return
        src = self.fmt_url(data.src, FileStatusType.FILE)
        dst = self.fmt_url(data.dst, FileStatusType.FILE)
        self._root.print(f"{src} -> {dst}")

    def leave(self, data: StorageProgressLeaveDir) -> None:
        pass

    def fail(self, data: StorageProgressFail) -> None:
        src = self.fmt_url(data.src, FileStatusType.FILE)
        dst = self.fmt_url(data.dst, FileStatusType.FILE)
        self._root.print(
            f"[red]Failure:[/] {src} -> {dst} \\[{data.message}]", err=True
        )


class TTYProgress(BaseStorageProgress):
    HEIGHT = 25
    FLUSH_INTERVAL = 0.2
    time_factory = staticmethod(monotonic)

    def __init__(
        self,
        root: Root,
        *,
        get_time: GetTimeCallable = time.monotonic,
        auto_refresh: bool = True,
    ) -> None:
        self.painter = get_painter(root.color)
        self._root = root
        self._mapping: Dict[URL, TaskID] = {}
        self._progress = Progress(
            TextColumn("[progress.description]{task.fields[filename]}"),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            DownloadColumn(),
            TransferSpeedColumn(),
            console=root.console,
            auto_refresh=auto_refresh,
            get_time=get_time,
        )
        self._progress.start()

    def _refresh(self) -> None:
        height = self._root.terminal_size[1]
        # If terminal is to small, we cannot show all tasks.
        # Moreover, scroll will be also broken, so the
        # only solution is just to hide some tasks.
        to_hide = len(self._progress.task_ids) - height + 2
        hidden = sum(1 for task in self._progress.tasks if not task.visible)
        for task in self._progress.tasks:
            if to_hide < hidden and not task.visible:
                self._progress.update(task.id, visible=True)
                hidden -= 1
            if to_hide > hidden and task.visible:
                self._progress.update(task.id, visible=False)
                hidden += 1
            if to_hide == hidden:
                break
        self._progress.refresh()

    def fmt_url(self, url: URL, type: FileStatusType) -> Text:
        return self.fmt_str(format_url(url), type)

    def fmt_str(self, label: str, type: FileStatusType) -> Text:
        return self.painter.paint(label, type)

    def begin(self, src: URL, dst: URL) -> StorageProgressContextManager:
        src_label = self.fmt_url(src, FileStatusType.DIRECTORY)
        dst_label = self.fmt_url(dst, FileStatusType.DIRECTORY)
        self._progress.log(Text.assemble("Copying ", src_label, " => ", dst_label))
        return super().begin(src, dst)

    def end(self) -> None:
        # Clean terminal if there is no files in progress
        self._progress.transient = not self._progress.task_ids
        self._progress.stop()

    def enter(self, data: StorageProgressEnterDir) -> None:
        fmt_src = self.fmt_url(data.src, FileStatusType.DIRECTORY)
        self._progress.log(Text.assemble("Starting copying directory ", fmt_src))

    def leave(self, data: StorageProgressLeaveDir) -> None:
        fmt_src = self.fmt_url(data.src, FileStatusType.DIRECTORY)
        self._progress.log(Text.assemble("Finished copying directory ", fmt_src))

    def start(self, data: StorageProgressStart) -> None:
        fmt_src = self.fmt_url(data.src, FileStatusType.FILE)
        fmt_name = self.fmt_str(data.src.name, FileStatusType.FILE)
        self._progress.log(Text.assemble("Copying: ", fmt_src))
        task_id = self._progress.add_task(
            description=data.src.name,
            total=data.size,
            filename=fmt_name,
        )
        self._mapping[data.src] = task_id
        self._refresh()

    def step(self, data: StorageProgressStep) -> None:
        task_id = self._mapping[data.src]
        self._progress.update(task_id, completed=data.current)
        self._refresh()

    def complete(self, data: StorageProgressComplete) -> None:
        fmt_src = self.fmt_url(data.src, FileStatusType.FILE)
        self._progress.log(Text.assemble("Copied: ", fmt_src))
        task_id = self._mapping[data.src]
        self._progress.remove_task(task_id)
        self._refresh()

    def fail(self, data: StorageProgressFail) -> None:
        src = self.fmt_str(str(data.src), FileStatusType.FILE)
        dst = self.fmt_str(str(data.dst), FileStatusType.FILE)
        self._root.print(
            f"[red]Failure:[/red] {src} -> {dst} \\[{data.message}]", err=True
        )


@dataclass(frozen=True)
class Tree:
    name: str
    size: int
    folders: Sequence["Tree"]
    files: Sequence[FileStatus]


class TreeFormatter:
    ANSI_DELIMS = ["├", "└", "─", "│"]
    SIMPLE_DELIMS = ["+", "+", "-", "|"]

    def __init__(
        self, *, color: bool, size: bool, human_readable: bool, sort: str
    ) -> None:
        self._ident: List[bool] = []
        self._numdirs = 0
        self._numfiles = 0
        self._painter = get_painter(color)
        if sys.platform != "win32":
            self._delims = self.ANSI_DELIMS
        else:
            self._delims = self.SIMPLE_DELIMS
        if human_readable:
            self._size_func = self._human_readable
        elif size:
            self._size_func = self._size
        else:
            self._size_func = self._none
        self._key = FilesSorter(sort).key()

    def __call__(self, tree: Tree) -> Text:
        ret = self.listdir(tree)
        ret.append(f"\n{self._numdirs} directories, {self._numfiles} files")
        return ret

    def listdir(self, tree: Tree) -> Text:
        ret = Text()
        items = sorted(tree.folders + tree.files, key=self._key)  # type: ignore
        ret.append(
            Text.assemble(
                self.pre(),
                self._size_func(tree.size),
                self._painter.paint(tree.name, FileStatusType.DIRECTORY),
                "\n",
            )
        )
        for num, item in enumerate(items):
            if isinstance(item, Tree):
                self._numdirs += 1
                with self.ident(num == len(items) - 1):
                    ret.append(self.listdir(item))
            else:
                self._numfiles += 1
                with self.ident(num == len(items) - 1):
                    ret.append(
                        Text.assemble(
                            self.pre(),
                            self._size_func(item.size),
                            self._painter.paint(item.name, FileStatusType.FILE),
                            "\n",
                        )
                    )
        return ret

    def pre(self) -> str:
        ret = []
        for last in self._ident[:-1]:
            if last:
                ret.append(" " * 4)
            else:
                ret.append(self._delims[3] + " " * 3)
        if self._ident:
            last = self._ident[-1]
            ret.append(self._delims[1] if last else self._delims[0])
            ret.append(self._delims[2] * 2)
            ret.append(" ")
        return "".join(ret)

    @contextlib.contextmanager
    def ident(self, last: bool) -> Iterator[None]:
        self._ident.append(last)
        try:
            yield
        finally:
            self._ident.pop()

    def _size(self, size: int) -> str:
        return f"[{size:>11}]  "

    def _human_readable(self, size: int) -> str:
        return f"[{format_size(size):>7}]  "

    def _none(self, size: int) -> str:
        return ""



import asyncio
import dataclasses
import io
import logging
from collections import namedtuple
from difflib import ndiff
from pathlib import Path
from typing import Any, AsyncIterator, Callable, DefaultDict, List, Optional, Set, Union

import click
import pytest
from rich.console import Console, RenderableType
from yarl import URL

import neuromation
from neuromation.api import Cluster, Factory, Preset
from neuromation.api.config import _AuthConfig, _AuthToken, _ConfigData
from neuromation.cli import main
from neuromation.cli.const import EX_OK
from neuromation.cli.root import Root


SysCapWithCode = namedtuple("SysCapWithCode", ["out", "err", "code"])
log = logging.getLogger(__name__)


@pytest.fixture()
def nmrc_path(tmp_path: Path, token: str, auth_config: _AuthConfig) -> Path:
    nmrc_path = tmp_path / "conftest.nmrc"
    cluster_config = Cluster(
        registry_url=URL("https://registry-dev.neu.ro"),
        storage_url=URL("https://storage-dev.neu.ro"),
        users_url=URL("https://users-dev.neu.ro"),
        monitoring_url=URL("https://monitoring-dev.neu.ro"),
        secrets_url=URL("https://secrets-dev.neu.ro"),
        presets={
            "gpu-small": Preset(
                cpu=7, memory_mb=30 * 1024, gpu=1, gpu_model="nvidia-tesla-k80"
            ),
            "gpu-large": Preset(
                cpu=7, memory_mb=60 * 1024, gpu=1, gpu_model="nvidia-tesla-v100"
            ),
            "cpu-small": Preset(cpu=7, memory_mb=2 * 1024),
            "cpu-large": Preset(cpu=7, memory_mb=14 * 1024),
        },
        name="default",
    )
    config = _ConfigData(
        auth_config=auth_config,
        auth_token=_AuthToken.create_non_expiring(token),
        url=URL("https://dev.neu.ro/api/v1"),
        version=neuromation.__version__,
        cluster_name="default",
        clusters={cluster_config.name: cluster_config},
    )
    Factory(nmrc_path)._save(config)
    return nmrc_path


def create_root(config_path: Path) -> Root:
    return Root(
        color=False,
        tty=False,
        disable_pypi_version_check=True,
        network_timeout=60,
        config_path=config_path,
        verbosity=0,
        trace=False,
        trace_hide_token=True,
        command_path="",
        command_params=[],
        skip_gmp_stats=True,
        show_traceback=False,
    )


@pytest.fixture()
async def root(nmrc_path: Path, loop: asyncio.AbstractEventLoop) -> AsyncIterator[Root]:
    root = create_root(config_path=nmrc_path)
    await root.init_client()
    yield root
    await root.client.close()


@pytest.fixture()
async def root_no_logged_in(
    tmp_path: Path, loop: asyncio.AbstractEventLoop
) -> AsyncIterator[Root]:
    root = create_root(config_path=tmp_path)
    assert root._client is None
    yield root
    assert root._client is None


@pytest.fixture()
def run_cli(
    nmrc_path: Path, capfd: Any, tmp_path: Path
) -> Callable[[List[str]], SysCapWithCode]:
    def _run_cli(arguments: List[str]) -> SysCapWithCode:
        log.info("Run 'neuro %s'", " ".join(arguments))
        capfd.readouterr()

        code = EX_OK
        try:
            default_args = [
                "--show-traceback",
                "--disable-pypi-version-check",
                "--color=no",
            ]
            if "--neuromation-config" not in arguments:
                for arg in arguments:
                    if arg.startswith("--neuromation-config="):
                        break
                else:
                    default_args.append(f"--neuromation-config={nmrc_path}")

            main(default_args + arguments)
        except SystemExit as e:
            code = e.code
            pass
        out, err = capfd.readouterr()
        return SysCapWithCode(out.strip(), err.strip(), code)

    return _run_cli


@pytest.fixture()
def click_tty_emulation(monkeypatch: Any) -> None:
    monkeypatch.setattr("click._compat.isatty", lambda stream: True)


@dataclasses.dataclass(eq=False)
class Guard:
    arg: str
    path: Path

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Guard):
            return NotImplemented
        return self.arg == other.arg


class RichComparator:
    def __init__(self, config: Any) -> None:
        self._regen = config.getoption("--rich-gen")
        self._config = config
        self._reporter = config.pluginmanager.getplugin("terminalreporter")
        assert self._reporter is not None
        self._cwd = Path.cwd()
        self._written_refs: List[Path] = []
        self._checked_refs: Set[Path] = set()
        self._file_pos = DefaultDict[io.StringIO, int](int)

    def mkref(self, request: Any, index: Optional[int]) -> Path:
        folder = Path(request.fspath).parent
        basename = request.function.__qualname__
        if hasattr(request.node, "callspec"):
            parametrize_id = request.node.callspec.id
            # Some characters are forbidden in FS path (on Windows)
            bad_to_good = {
                "/": "#forward_slash#",
                "\\": "#back_slash#",
                "<": "#less#",
                ">": "#more#",
                ":": "#colon#",
                '"': "#double_qoute#",
                "|": "#vertical_bar#",
                "?": "#question_mark#",
                "*": "#star#",
            }
            for bad, good in bad_to_good.items():
                parametrize_id = parametrize_id.replace(bad, good)
            # On windows, some characters are forbidden
            basename += f"[{parametrize_id}]"
        if index is not None:
            basename += "_" + str(index)
        basename += ".ref"
        return folder / "ascii" / basename

    def rel(self, ref: Path) -> Path:
        return ref.relative_to(self._cwd)

    def check_io(self, ref: Path, file: io.StringIO) -> None:
        __tracebackhide__ = True
        tmp = file.getvalue()
        buf = tmp[self._file_pos[file] :]
        self._file_pos[file] = len(tmp)
        self.check(ref, buf)

    def check(self, ref: Path, buf: str) -> None:
        __tracebackhide__ = True

        if ref in self._checked_refs:
            pytest.fail(
                f"{self.rel(ref)} is already checked. "
                "Hint: use index when generating refs automatically"
            )
        else:
            self._checked_refs.add(ref)

        buf = buf.strip()
        buf = click.unstyle(buf)

        if self._regen:
            self.write_ref(ref, buf)
        else:
            orig = self.read_ref(ref)
            tmp = ref.with_suffix(".orig")
            self.write_file(tmp, buf)
            # reading from file is important, fil writer replaces \r with \n
            actual = self.read_file(tmp)
            assert Guard(actual, tmp) == Guard(orig, ref)

    def read_file(self, ref: Path) -> str:
        return ref.read_text(encoding="utf8").strip()

    def read_ref(self, ref: Path) -> str:
        __tracebackhide__ = True
        if not ref.exists():
            rel_ref = self.rel(ref)
            pytest.fail(
                f"The reference {rel_ref} doesn't exist.\n"
                "Create it yourself or run pytest with '--rich-gen' option."
            )
        else:
            return self.read_file(ref)

    def write_file(self, ref: Path, buf: str) -> None:
        ref.parent.mkdir(parents=True, exist_ok=True)
        ref.write_text(buf.strip() + "\n", encoding="utf8")

    def write_ref(self, ref: Path, buf: str) -> bool:
        if ref.exists():
            orig = ref.read_text().strip()
            if orig == buf:
                return False
        self.write_file(ref, buf)
        if self._reporter.verbosity > 0:
            rel_ref = self.rel(ref)
            self._reporter.write_line(f"Regenerate {rel_ref}", yellow=True)
        self._written_refs.append(ref)
        return True

    def summary(self) -> None:
        if self._reporter.verbosity == 0:
            if self._written_refs:
                self._reporter.write_line("Regenerated files:", yellow=True)
                for fname in self._written_refs:
                    rel_ref = self.rel(fname)
                    self._reporter.write_line(f"  {rel_ref}", yellow=True)

    def diff(self, lft: Guard, rgt: Guard) -> List[str]:
        # The same as _diff_text from
        # https://github.com/pytest-dev/pytest/blob/master/src/_pytest/assertion/util.py#L200-L245 plus a few extra lines with additional instructions.  # noqa
        explanation: List[str] = []

        left = lft.arg
        right = rgt.arg

        if self._reporter.verbosity < 1:
            i = 0  # just in case left or right has zero length
            for i in range(min(len(left), len(right))):
                if left[i] != right[i]:
                    break
            if i > 42:
                i -= 10  # Provide some context
                explanation = [
                    "Skipping %s identical leading characters in diff, use -v to show"
                    % i
                ]
                left = left[i:]
                right = right[i:]
            if len(left) == len(right):
                for i in range(len(left)):
                    if left[-i] != right[-i]:
                        break
                if i > 42:
                    i -= 10  # Provide some context
                    explanation += [
                        "Skipping {} identical trailing "
                        "characters in diff, use -v to show".format(i)
                    ]
                    left = left[:-i]
                    right = right[:-i]

        keepends = True
        if left.isspace() or right.isspace():
            left = repr(str(left))
            right = repr(str(right))
            explanation += [
                "Strings contain only whitespace, escaping them using repr()"
            ]
        # "right" is the expected base against which we compare "left",
        # see https://github.com/pytest-dev/pytest/issues/3333
        explanation += [
            line.strip("\n")
            for line in ndiff(right.splitlines(keepends), left.splitlines(keepends))
        ]
        explanation.append("")
        explanation.append(f"'cat {self.rel(lft.path)}' to see the test output.")
        explanation.append(f"'cat {self.rel(rgt.path)}' to see the reference.")
        explanation.append(
            f"Use 'pytest ... --rich-gen' to regenerate reference files "
            "from values calculated by tests"
        )
        return explanation


def pytest_assertrepr_compare(
    config: Any, op: str, left: object, right: object
) -> Optional[List[str]]:
    if isinstance(left, Guard) and isinstance(right, Guard):
        plugin = config.pluginmanager.getplugin("rich-comparator")
        return plugin.diff(left, right)
    return None


# run after terminalreporter/capturemanager are configured
@pytest.hookimpl(trylast=True)
def pytest_configure(config: Any) -> None:
    comparator = RichComparator(config)
    config.pluginmanager.register(comparator, "rich-comparator")


def pytest_terminal_summary(terminalreporter: Any) -> None:
    config = terminalreporter.config
    comparator = config.pluginmanager.getplugin("rich-comparator")
    comparator.summary()


@pytest.fixture
def rich_cmp(request: Any) -> Callable[..., None]:
    def comparator(
        src: Union[RenderableType, Console],
        ref: Optional[Path] = None,
        *,
        color: bool = True,
        tty: bool = True,
        index: Optional[int] = 0,
    ) -> None:
        __tracebackhide__ = True
        plugin = request.config.pluginmanager.getplugin("rich-comparator")
        if ref is None:
            ref = plugin.mkref(request, index)

        if isinstance(src, io.StringIO):
            plugin.check_io(ref, src)
        elif isinstance(src, Console):
            if isinstance(src.file, io.StringIO):
                plugin.check_io(ref, src.file)
            else:
                buf = src.export_text(clear=True, styles=True)
                plugin.check(ref, buf)
        else:
            file = io.StringIO()
            console = Console(
                file=file,
                width=160,
                height=24,
                force_terminal=tty,
                color_system="auto" if color else None,
                record=True,
                highlighter=None,
                legacy_windows=False,
            )
            console.print(src)
            plugin.check_io(ref, file)

    return comparator


NewConsole = Callable[..., Console]


@pytest.fixture
def new_console() -> NewConsole:
    def factory(*, tty: bool, color: bool = True) -> Console:
        file = io.StringIO()
        # console doesn't accept the time source,
        # using the real time in tests is not reliable
        return Console(
            file=file,
            width=160,
            height=24,
            force_terminal=tty,
            color_system="auto" if color else None,
            record=True,
            highlighter=None,
            legacy_windows=False,
            log_path=False,
            log_time=False,
        )

    return factory



import asyncio
import re
import subprocess
import time
from dataclasses import replace
from pathlib import Path
from typing import Any, AsyncIterator, Set
from uuid import uuid4 as uuid

import aiodocker
import pytest
from yarl import URL

from neuromation.api import CONFIG_ENV_NAME, DEFAULT_CONFIG_PATH, JobStatus
from tests.e2e import Helper


TEST_IMAGE_NAME = "e2e-banana-image"


def parse_docker_ls_output(docker_ls_output: Any) -> Set[str]:
    return {
        repo_tag
        for info in docker_ls_output
        if info["RepoTags"] is not None
        for repo_tag in info["RepoTags"]
        if repo_tag
    }


@pytest.fixture()
def tag() -> str:
    return str(uuid())


async def generate_image(
    docker: aiodocker.Docker, tag: str, name: str = TEST_IMAGE_NAME
) -> str:
    image_archive = Path(__file__).parent / "assets/echo-tag.tar"
    # TODO use random image name here
    image_name = f"{name}:{tag}"
    with image_archive.open(mode="r+b") as fileobj:
        await docker.images.build(
            fileobj=fileobj, tag=image_name, buildargs={"TAG": tag}, encoding="identity"
        )

    return image_name


@pytest.fixture()
async def image(
    loop: asyncio.AbstractEventLoop, docker: aiodocker.Docker, tag: str
) -> AsyncIterator[str]:
    image = await generate_image(docker, tag)
    yield image
    await docker.images.delete(image, force=True)


@pytest.mark.e2e
def test_images_complete_lifecycle(
    helper: Helper,
    image: str,
    tag: str,
    loop: asyncio.AbstractEventLoop,
    docker: aiodocker.Docker,
) -> None:
    # Let`s push image
    captured = helper.run_cli(["image", "push", image])

    # stderr has "Used image ..." lines
    # assert not captured.err

    image_full_str = f"image://{helper.cluster_name}/{helper.username}/{image}"
    assert captured.out.endswith(image_full_str)
    image_url = URL(image_full_str)

    # Check if image available on registry
    captured = helper.run_cli(["image", "ls", "--full-uri"])

    image_urls = [URL(line.strip()) for line in captured.out.splitlines()[2:] if line]
    for url in image_urls:
        assert url.scheme == "image", url
    image_url_without_tag = image_url.with_path(image_url.path.replace(f":{tag}", ""))
    assert image_url_without_tag in image_urls

    # delete local
    loop.run_until_complete(docker.images.delete(image, force=True))
    docker_ls_output = loop.run_until_complete(docker.images.list())
    local_images = parse_docker_ls_output(docker_ls_output)
    assert image not in local_images

    # Pull image as with another tag
    captured = helper.run_cli(["image", "pull", f"image:{image}"])
    # stderr has "Used image ..." lines
    # assert not captured.err
    assert captured.out.endswith(image)

    # check pulled locally, delete for cleanup
    docker_ls_output = loop.run_until_complete(docker.images.list())
    local_images = parse_docker_ls_output(docker_ls_output)
    assert image in local_images

    # Execute image and check result
    captured = helper.run_cli(["-q", "run", "--no-wait-start", str(image_url)])
    assert not captured.err
    job_id = captured.out
    assert job_id.startswith("job-")
    helper.wait_job_change_state_to(job_id, JobStatus.SUCCEEDED, JobStatus.FAILED)

    helper.check_job_output(job_id, re.escape(tag))


@pytest.mark.e2e
def test_image_tags(helper: Helper, image: str, tag: str) -> None:
    # push image
    captured = helper.run_cli(["image", "push", image])

    image_full_str = f"image://{helper.cluster_name}/{helper.username}/{image}"
    assert captured.out.endswith(image_full_str)

    image_full_str_no_tag = image_full_str.replace(f":{tag}", "")

    delay = 0
    t0 = time.time()

    while time.time() - t0 < 180:
        time.sleep(delay)
        # check the tag is present now
        captured = helper.run_cli(["image", "tags", image_full_str_no_tag])
        if tag in map(lambda s: s.strip(), captured.out.splitlines()):
            break
        # Give a chance to sync remote registries
        delay = min(delay * 2 + 1, 15)
    else:
        raise AssertionError(
            f"Delay is reached on waiting for tag {tag} in {captured.out}"
        )

    cmd = f"neuro image tags {image_full_str}"
    result = subprocess.run(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    assertion_msg = f"Command {cmd} should fail: {result.stdout!r} {result.stderr!r}"
    assert result.returncode, assertion_msg

    image_full_str_latest_tag = image_full_str.replace(f":{tag}", ":latest")
    cmd = f"neuro image tags {image_full_str_latest_tag}"
    result = subprocess.run(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    assertion_msg = f"Command {cmd} should fail: {result.stdout!r} {result.stderr!r}"
    assert result.returncode, assertion_msg


@pytest.mark.e2e
def test_image_ls(helper: Helper, image: str, tag: str) -> None:
    # push image
    captured = helper.run_cli(["image", "push", image])

    image_full_str = f"image://{helper.cluster_name}/{helper.username}/{image}"
    image_short_str = f"image:{image}"
    assert captured.out.endswith(image_full_str)

    image_full_str_no_tag = image_full_str.replace(f":{tag}", "")
    image_short_str_no_tag = image_short_str.replace(f":{tag}", "")

    # check ls short mode
    captured = helper.run_cli(["image", "ls"])
    assert image_short_str_no_tag in [
        line.strip() for line in captured.out.splitlines()
    ]

    captured = helper.run_cli(["image", "ls", "--full-uri"])
    assert image_full_str_no_tag in [line.strip() for line in captured.out.splitlines()]

    # check ls long mode
    captured = helper.run_cli(["image", "ls", "-l"])
    for line in captured.out.splitlines():
        if image_short_str_no_tag in line:
            break
    else:
        assert False, f"Not found {image_short_str_no_tag} in {captured.out}"


@pytest.mark.e2e
async def test_images_delete(
    helper: Helper,
    docker: aiodocker.Docker,
) -> None:
    name = f"test-{uuid()}"
    await generate_image(docker, name=name, tag="latest")
    img_name = f"image:{name}"

    helper.run_cli(["image", "push", name + ":latest"])

    captured = helper.run_cli(["-q", "image", "ls"])
    assert img_name in captured.out

    helper.run_cli(["image", "rm", img_name])

    for _ in range(10):
        captured = helper.run_cli(["-q", "image", "ls"])
        if img_name in captured.out:
            time.sleep(5)

    assert img_name not in captured.out


@pytest.mark.e2e
async def test_images_push_with_specified_name(
    helper: Helper,
    image: str,
    tag: str,
    loop: asyncio.AbstractEventLoop,
    docker: aiodocker.Docker,
) -> None:
    # Let`s push image
    image_no_tag = image.replace(f":{tag}", "")
    pushed_no_tag = f"{image_no_tag}-pushed"
    pulled_no_tag = f"{image_no_tag}-pulled"
    pulled = f"{pulled_no_tag}:{tag}"

    captured = helper.run_cli(["image", "push", image, f"image:{pushed_no_tag}:{tag}"])
    # stderr has "Used image ..." lines
    # assert not captured.err
    async with helper.client() as client:
        image_pushed_full_str = (
            f"image://{client.config.cluster_name}/"
            f"{client.config.username}/{pushed_no_tag}:{tag}"
        )
        assert captured.out.endswith(image_pushed_full_str)

    # Check if image available on registry
    docker_ls_output = await docker.images.list()
    local_images = parse_docker_ls_output(docker_ls_output)
    assert pulled not in local_images

    async with helper.client() as client:
        image_pushed_full = client.parse.remote_image(image_pushed_full_str)
        image_url_without_tag = replace(image_pushed_full, tag=None)
        imgs = await client.images.ls()
        assert image_url_without_tag in imgs

    # check locally
    docker_ls_output = await docker.images.list()
    local_images = parse_docker_ls_output(docker_ls_output)
    assert pulled not in local_images

    # Pull image as with another name
    captured = helper.run_cli(["image", "pull", f"image:{pushed_no_tag}:{tag}", pulled])
    # stderr has "Used image ..." lines
    # assert not captured.err
    assert captured.out.endswith(pulled)
    # check locally
    docker_ls_output = await docker.images.list()
    local_images = parse_docker_ls_output(docker_ls_output)
    assert pulled in local_images

    # TODO (A.Yushkovskiy): delete the pushed image in GCR
    # delete locally
    await docker.images.delete(pulled, force=True)


@pytest.mark.e2e
def test_docker_helper(
    helper: Helper, image: str, tag: str, nmrc_path: Path, monkeypatch: Any
) -> None:
    monkeypatch.setenv(CONFIG_ENV_NAME, str(nmrc_path or DEFAULT_CONFIG_PATH))
    helper.run_cli(["config", "docker"])
    registry = helper.registry_url.host
    username = helper.username
    full_tag = f"{registry}/{username}/{image}"
    tag_cmd = f"docker tag {image} {full_tag}"
    result = subprocess.run(
        tag_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    assert (
        not result.returncode
    ), f"Command {tag_cmd} failed: {result.stdout!r} {result.stderr!r} "
    push_cmd = f"docker push {full_tag}"
    result = subprocess.run(
        push_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    assert (
        not result.returncode
    ), f"Command {push_cmd} failed: {result.stdout!r} {result.stderr!r} "
    # Run image and check output
    image_url = f"image://{helper.cluster_name}/{username}/{image}"
    job_id = helper.run_job_and_wait_state(
        image_url, "", wait_state=JobStatus.SUCCEEDED, stop_state=JobStatus.FAILED
    )
    helper.check_job_output(job_id, re.escape(tag))



from typing import Any

import pytest

from neuromation.cli.printer import StreamPrinter, TTYPrinter


class TestStreamPrinter:
    @pytest.fixture
    def printer(self) -> StreamPrinter:
        return StreamPrinter()

    def test_no_messages(self, printer: StreamPrinter, capfd: Any) -> None:
        printer.close()
        out, err = capfd.readouterr()
        assert err == ""
        assert out == ""

    def test_one_message(self, printer: StreamPrinter, capfd: Any) -> None:
        printer.print("message")
        printer.close()
        out, err = capfd.readouterr()
        assert err == ""
        assert out == "message\n"

    def test_two_messages(self, printer: StreamPrinter, capfd: Any) -> None:
        printer.print("message1")
        printer.print("message2")
        printer.close()
        out, err = capfd.readouterr()
        assert err == ""
        assert out == "message1\nmessage2\n"

    def test_ticks_without_messages(self, printer: StreamPrinter, capfd: Any) -> None:
        printer.tick()
        printer.close()
        out, err = capfd.readouterr()
        assert err == ""
        assert out == ".\n"

    def test_ticks_with_messages(
        self, printer: StreamPrinter, capfd: Any, monkeypatch: Any
    ) -> None:
        monkeypatch.setattr("neuromation.cli.printer.TICK_TIMEOUT", 0)
        printer.tick()
        printer.print("message")
        printer.tick()
        printer.tick()
        printer.close()
        out, err = capfd.readouterr()
        assert err == ""
        assert out == ".\nmessage..\n"

    def test_ticks_spam_control(
        self, printer: StreamPrinter, capfd: Any, monkeypatch: Any
    ) -> None:
        monkeypatch.setattr("neuromation.cli.printer.TICK_TIMEOUT", 1000)
        printer.tick()
        printer.tick()
        printer.close()
        out, err = capfd.readouterr()
        assert err == ""
        assert out == ".\n"


class TestTTYPrinter:
    @pytest.fixture
    def printer(self, click_tty_emulation: Any) -> TTYPrinter:
        return TTYPrinter()

    def test_no_messages(self, capfd: Any, printer: TTYPrinter) -> None:
        printer.close()
        out, err = capfd.readouterr()
        assert err == ""
        assert out == ""

    def test_one_message(self, capfd: Any, printer: TTYPrinter) -> None:
        printer.print("message")
        printer.close()
        out, err = capfd.readouterr()
        assert err == ""
        assert out == "message\n"

    def test_two_messages(self, capfd: Any, printer: TTYPrinter) -> None:
        printer.print("message1")
        printer.print("message2")
        printer.close()
        out, err = capfd.readouterr()
        assert err == ""
        assert out == "message1\nmessage2\n"

    # very simple test
    def test_message_lineno(self, printer: TTYPrinter, capfd: Any) -> None:
        assert printer.total_lines == 0
        printer.print("message1")
        assert printer.total_lines == 1
        printer.print("message1-replace", 0)
        assert printer.total_lines == 1
        printer.print("message3", 2)
        assert printer.total_lines == 3
        printer.print("message7", 6)
        assert printer.total_lines == 7
        printer.print("message2", 1)
        assert printer.total_lines == 7
        printer.close()
        out, err = capfd.readouterr()
        assert err == ""
        assert "message1" in out
        assert "message1-replace" in out
        assert "message3" in out
        assert "message7" in out
        assert "message2" in out
        CSI = "\033["
        assert CSI in out
        assert f"{CSI}0A" not in out
        assert f"{CSI}0B" not in out

    def test_multiline(self, printer: TTYPrinter, capfd: Any) -> None:
        CSI = "\033["
        assert printer.total_lines == 0
        res = printer.print("message1\nmessage2")
        assert printer.total_lines == 2
        assert res == "message1\nmessage2\n"

        res = printer.print("message3\nmessage4", 1)
        assert printer.total_lines == 3
        assert res == f"{CSI}1Amessage3{CSI}0K\nmessage4\n"

        res = printer.print("message5\nmessage6", 10)
        assert printer.total_lines == 12
        assert res == f"\n\n\n\n\n\n\n{CSI}1Amessage5\nmessage6\n"

        res = printer.print("message7\nmessage8", 5)
        assert printer.total_lines == 12
        assert res == f"{CSI}7Amessage7{CSI}0K\nmessage8{CSI}0K\n{CSI}5B"



from datetime import timedelta
from pathlib import Path
from typing import Any, Callable, Dict, NoReturn
from unittest import mock

import click
import pytest
import toml
from aiohttp import web
from yarl import URL

from neuromation.api import Action, Client, JobStatus
from neuromation.cli.parse_utils import parse_timedelta
from neuromation.cli.root import Root
from neuromation.cli.utils import (
    calc_life_span,
    pager_maybe,
    parse_file_resource,
    parse_permission_action,
    parse_resource_for_sharing,
    resolve_job,
)
from tests import _TestServerFactory


_MakeClient = Callable[..., Client]


def _job_entry(job_id: str) -> Dict[str, Any]:
    return {
        "id": job_id,
        "owner": "job-owner",
        "cluster_name": "default",
        "uri": f"job://default/job-owner/{job_id}",
        "status": "running",
        "history": {
            "status": "running",
            "reason": None,
            "description": None,
            "created_at": "2019-03-18T12:41:10.573468+00:00",
            "started_at": "2019-03-18T12:41:16.804040+00:00",
        },
        "container": {
            "image": "ubuntu:latest",
            "env": {},
            "volumes": [],
            "command": "sleep 1h",
            "resources": {"cpu": 0.1, "memory_mb": 1024, "shm": True},
        },
        "ssh_auth_server": "ssh://nobody@ssh-auth-dev.neu.ro:22",
        "is_preemptible": True,
        "pass_config": False,
        "name": "job-name",
        "internal_hostname": "job-id.default",
        "internal_hostname_named": "job-name--job-owner.default",
    }


async def test_resolve_job_id__from_string__no_jobs_found(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    JSON: Dict[str, Any] = {"jobs": []}
    job_id = "job-81839be3-3ecf-4ec5-80d9-19b1588869db"

    async def handler(request: web.Request) -> web.Response:
        # Since `resolve_job` excepts any Exception, `assert` will be caught there
        name = request.query.get("name")
        if name != job_id:
            raise web.HTTPBadRequest(text=(f"received: name={name}"))
        owner = request.query.get("owner")
        if owner != "user":
            raise web.HTTPBadRequest(text=(f"received: owner={owner}"))
        reverse = request.query.get("reverse")
        if reverse != "1":
            raise web.HTTPBadRequest(text=(f"received: reverse={reverse}"))
        limit = request.query.get("limit")
        if limit != "1":
            raise web.HTTPBadRequest(text=(f"received: limit={limit}"))
        status = request.query.getall("status")
        if status != ["running"]:
            raise web.HTTPBadRequest(text=(f"received: status={status}"))
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get("/jobs", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        resolved = await resolve_job(job_id, client=client, status={JobStatus.RUNNING})
        assert resolved == job_id


async def test_resolve_job_id__from_uri_with_owner_same_user__no_jobs_found(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    job_owner = "user"
    job_name = "job-name"
    uri = f"job://default/{job_owner}/{job_name}"
    JSON: Dict[str, Any] = {"jobs": []}

    async def handler(request: web.Request) -> web.Response:
        # Since `resolve_job` excepts any Exception, `assert` will be caught there
        name = request.query.get("name")
        if name != job_name:
            pytest.fail(f"received: {name}")
        owner = request.query.get("owner")
        if owner != job_owner:
            pytest.fail(f"received: {owner}")
        reverse = request.query.get("reverse")
        if reverse != "1":
            pytest.fail(f"received: reverse={reverse}")
        limit = request.query.get("limit")
        if limit != "1":
            pytest.fail(f"received: limit={limit}")
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get("/jobs", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        resolved = await resolve_job(uri, client=client, status={JobStatus.RUNNING})
        assert resolved == job_name


async def test_resolve_job_id__from_uri_with_owner_other_user__no_jobs_found(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    job_owner = "job-owner"
    job_name = "job-name"
    uri = f"job://default/{job_owner}/{job_name}"
    JSON: Dict[str, Any] = {"jobs": []}

    async def handler(request: web.Request) -> web.Response:
        # Since `resolve_job` excepts any Exception, `assert` will be caught there
        name = request.query.get("name")
        if name != job_name:
            pytest.fail(f"received: {name}")
        owner = request.query.get("owner")
        if owner != job_owner:
            pytest.fail(f"received: {owner}")
        reverse = request.query.get("reverse")
        if reverse != "1":
            pytest.fail(f"received: reverse={reverse}")
        limit = request.query.get("limit")
        if limit != "1":
            pytest.fail(f"received: limit={limit}")
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get("/jobs", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        with pytest.raises(ValueError):
            await resolve_job(uri, client=client, status={JobStatus.RUNNING})


async def test_resolve_job_id__from_uri_without_owner__no_jobs_found(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    job_name = "job-name"
    uri = f"job:{job_name}"
    JSON: Dict[str, Any] = {"jobs": []}

    async def handler(request: web.Request) -> web.Response:
        # Since `resolve_job` excepts any Exception, `assert` will be caught there
        name = request.query.get("name")
        if name != job_name:
            pytest.fail(f"received: {name}")
        owner = request.query.get("owner")
        if owner != "user":
            pytest.fail(f"received: {owner}")
        reverse = request.query.get("reverse")
        if reverse != "1":
            pytest.fail(f"received: reverse={reverse}")
        limit = request.query.get("limit")
        if limit != "1":
            pytest.fail(f"received: limit={limit}")
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get("/jobs", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        resolved = await resolve_job(uri, client=client, status={JobStatus.RUNNING})
        assert resolved == job_name


async def test_resolve_job_id__from_string__single_job_found(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    job_name = "test-job-name-555"
    job_id = "job-id-1"
    JSON = {"jobs": [_job_entry(job_id)]}

    async def handler(request: web.Request) -> web.Response:
        # Since `resolve_job` excepts any Exception, `assert` will be caught there
        name = request.query.get("name")
        if name != job_name:
            pytest.fail(f"received: {name}")
        owner = request.query.get("owner")
        if owner != "user":
            pytest.fail(f"received: {owner}")
        reverse = request.query.get("reverse")
        if reverse != "1":
            pytest.fail(f"received: reverse={reverse}")
        limit = request.query.get("limit")
        if limit != "1":
            pytest.fail(f"received: limit={limit}")
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get("/jobs", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        resolved = await resolve_job(
            job_name, client=client, status={JobStatus.RUNNING}
        )
        assert resolved == job_id


async def test_resolve_job_id__from_uri_with_owner__single_job_found(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    job_owner = "job-owner"
    job_name = "job-name"
    uri = f"job://default/{job_owner}/{job_name}"
    job_id = "job-id-1"
    JSON = {"jobs": [_job_entry(job_id)]}

    async def handler(request: web.Request) -> web.Response:
        # Since `resolve_job` excepts any Exception, `assert` will be caught there
        name = request.query.get("name")
        if name != job_name:
            pytest.fail(f"received: {name}")
        owner = request.query.get("owner")
        if owner != job_owner:
            pytest.fail(f"received: {owner}")
        reverse = request.query.get("reverse")
        if reverse != "1":
            pytest.fail(f"received: reverse={reverse}")
        limit = request.query.get("limit")
        if limit != "1":
            pytest.fail(f"received: limit={limit}")
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get("/jobs", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        resolved = await resolve_job(uri, client=client, status={JobStatus.RUNNING})
        assert resolved == job_id


async def test_resolve_job_id__from_uri_without_owner__single_job_found(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    job_name = "job-name"
    uri = f"job:{job_name}"
    job_id = "job-id-1"
    JSON = {"jobs": [_job_entry(job_id)]}

    async def handler(request: web.Request) -> web.Response:
        # Since `resolve_job` excepts any Exception, `assert` will be caught there
        name = request.query.get("name")
        if name != job_name:
            pytest.fail(f"received: {name}")
        owner = request.query.get("owner")
        if owner != "user":
            pytest.fail(f"received: {owner}")
        reverse = request.query.get("reverse")
        if reverse != "1":
            pytest.fail(f"received: reverse={reverse}")
        limit = request.query.get("limit")
        if limit != "1":
            pytest.fail(f"received: limit={limit}")
        return web.json_response(JSON)

    app = web.Application()
    app.router.add_get("/jobs", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        resolved = await resolve_job(uri, client=client, status={JobStatus.RUNNING})
        assert resolved == job_id


async def test_resolve_job_id__server_error(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    job_id = "job-81839be3-3ecf-4ec5-80d9-19b1588869db"
    job_name = job_id

    async def handler(request: web.Request) -> NoReturn:
        # Since `resolve_job` excepts any Exception, `assert` will be caught there
        name = request.query.get("name")
        if name != job_name:
            pytest.fail(f"received: {name}")
        owner = request.query.get("owner")
        if owner != "user":
            pytest.fail(f"received: {owner}")
        reverse = request.query.get("reverse")
        if reverse != "1":
            pytest.fail(f"received: reverse={reverse}")
        limit = request.query.get("limit")
        if limit != "1":
            pytest.fail(f"received: limit={limit}")
        raise web.HTTPError()

    app = web.Application()
    app.router.add_get("/jobs", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        resolved = await resolve_job(
            job_name, client=client, status={JobStatus.RUNNING}
        )
        assert resolved == job_id


async def test_resolve_job_id__from_uri_with_owner_same_user__server_error(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    job_owner = "user"
    job_name = "job-name"
    uri = f"job://default/{job_owner}/{job_name}"

    async def handler(request: web.Request) -> NoReturn:
        # Since `resolve_job` excepts any Exception, `assert` will be caught there
        name = request.query.get("name")
        if name != job_name:
            pytest.fail(f"received: {name}")
        owner = request.query.get("owner")
        if owner != job_owner:
            pytest.fail(f"received: {owner}")
        reverse = request.query.get("reverse")
        if reverse != "1":
            pytest.fail(f"received: reverse={reverse}")
        limit = request.query.get("limit")
        if limit != "1":
            pytest.fail(f"received: limit={limit}")
        raise web.HTTPError()

    app = web.Application()
    app.router.add_get("/jobs", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        resolved = await resolve_job(uri, client=client, status={JobStatus.RUNNING})
        assert resolved == job_name


async def test_resolve_job_id__from_uri_with_owner_other_user__server_error(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    job_owner = "job-owner"
    job_name = "job-name"
    uri = f"job://default/{job_owner}/{job_name}"

    async def handler(request: web.Request) -> NoReturn:
        # Since `resolve_job` excepts any Exception, `assert` will be caught there
        name = request.query.get("name")
        if name != job_name:
            pytest.fail(f"received: {name}")
        owner = request.query.get("owner")
        if owner != job_owner:
            pytest.fail(f"received: {owner}")
        reverse = request.query.get("reverse")
        if reverse != "1":
            pytest.fail(f"received: reverse={reverse}")
        limit = request.query.get("limit")
        if limit != "1":
            pytest.fail(f"received: limit={limit}")
        raise web.HTTPError()

    app = web.Application()
    app.router.add_get("/jobs", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        with pytest.raises(ValueError):
            await resolve_job(uri, client=client, status={JobStatus.RUNNING})


async def test_resolve_job_id__from_uri_without_owner__server_error(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:
    job_name = "job-name"
    uri = f"job:{job_name}"

    async def handler(request: web.Request) -> NoReturn:
        # Since `resolve_job` excepts any Exception, `assert` will be caught there
        name = request.query.get("name")
        if name != job_name:
            pytest.fail(f"received: {name}")
        owner = request.query.get("owner")
        if owner != "user":
            pytest.fail(f"received: {owner}")
        reverse = request.query.get("reverse")
        if reverse != "1":
            pytest.fail(f"received: reverse={reverse}")
        limit = request.query.get("limit")
        if limit != "1":
            pytest.fail(f"received: limit={limit}")
        raise web.HTTPError()

    app = web.Application()
    app.router.add_get("/jobs", handler)

    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        resolved = await resolve_job(uri, client=client, status={JobStatus.RUNNING})
        assert resolved == job_name


async def test_resolve_job_id__from_uri__missing_job_id(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:

    uri = "job://default/job-name"

    app = web.Application()
    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        with pytest.raises(
            ValueError,
            match="Invalid job URI: owner='job-name', missing job-id or job-name",
        ):
            await resolve_job(uri, client=client, status={JobStatus.RUNNING})


async def test_resolve_job_id__from_uri__missing_job_id_2(
    aiohttp_server: _TestServerFactory, make_client: _MakeClient
) -> None:

    uri = "job://job-name"

    app = web.Application()
    srv = await aiohttp_server(app)

    async with make_client(srv.make_url("/")) as client:
        with pytest.raises(
            ValueError,
            match="Invalid job URI: cluster_name != 'default'",
        ):
            await resolve_job(uri, client=client, status={JobStatus.RUNNING})


def test_parse_file_resource_no_scheme(root: Root) -> None:
    parsed = parse_file_resource("scheme-less/resource", root)
    assert parsed == URL((Path.cwd() / "scheme-less/resource").as_uri())
    parsed = parse_file_resource("c:scheme-less/resource", root)
    assert parsed == URL((Path("c:scheme-less").resolve() / "resource").as_uri())


def test_parse_file_resource_unsupported_scheme(root: Root) -> None:
    with pytest.raises(ValueError, match=r"Unsupported URI scheme"):
        parse_file_resource("http://neuromation.io", root)
    with pytest.raises(ValueError, match=r"Unsupported URI scheme"):
        parse_file_resource("image:ubuntu", root)


def test_parse_file_resource_user_less(root: Root) -> None:
    user_less_permission = parse_file_resource("storage:resource", root)
    assert user_less_permission == URL(
        f"storage://{root.client.cluster_name}/{root.client.username}/resource"
    )


def test_parse_file_resource_with_user(root: Root) -> None:
    full_permission = parse_file_resource(
        f"storage://{root.client.cluster_name}/{root.client.username}/resource", root
    )
    assert full_permission == URL(
        f"storage://{root.client.cluster_name}/{root.client.username}/resource"
    )
    full_permission = parse_file_resource(f"storage://default/alice/resource", root)
    assert full_permission == URL(f"storage://default/alice/resource")


def test_parse_file_resource_with_tilde(root: Root) -> None:
    with pytest.raises(ValueError, match=r"Cannot expand user for "):
        parse_file_resource(f"storage://~/resource", root)


def test_parse_resource_for_sharing_image_no_tag(root: Root) -> None:
    uri = "image:ubuntu"
    parsed = parse_resource_for_sharing(uri, root)
    assert parsed == URL(
        f"image://{root.client.cluster_name}/{root.client.username}/ubuntu"
    )


def test_parse_resource_for_sharing_image_with_tag_fail(root: Root) -> None:
    uri = "image://~/ubuntu:latest"
    with pytest.raises(ValueError, match="tag is not allowed"):
        parse_resource_for_sharing(uri, root)


def test_parse_resource_for_sharing_no_scheme(root: Root) -> None:
    with pytest.raises(ValueError, match=r"URI Scheme not specified"):
        parse_resource_for_sharing("scheme-less/resource", root)


def test_parse_resource_for_sharing_unsupported_scheme(root: Root) -> None:
    with pytest.raises(ValueError, match=r"Unsupported URI scheme"):
        parse_resource_for_sharing("http://neuromation.io", root)
    with pytest.raises(ValueError, match=r"Unsupported URI scheme"):
        parse_resource_for_sharing("file:///etc/password", root)
    with pytest.raises(ValueError, match=r"Unsupported URI scheme"):
        parse_resource_for_sharing(r"c:scheme-less/resource", root)


def test_parse_resource_for_sharing_user_less(root: Root) -> None:
    user_less_permission = parse_resource_for_sharing("storage:resource", root)
    assert user_less_permission == URL(
        f"storage://{root.client.cluster_name}/{root.client.username}/resource"
    )


def test_parse_resource_for_sharing_with_user(root: Root) -> None:
    full_permission = parse_resource_for_sharing(
        f"storage://{root.client.cluster_name}/{root.client.username}/resource", root
    )
    assert full_permission == URL(
        f"storage://{root.client.cluster_name}/{root.client.username}/resource"
    )
    full_permission = parse_resource_for_sharing(
        f"storage://default/alice/resource", root
    )
    assert full_permission == URL(f"storage://default/alice/resource")


def test_parse_resource_for_sharing_with_tilde(root: Root) -> None:
    with pytest.raises(ValueError, match=r"Cannot expand user for "):
        parse_resource_for_sharing(f"storage://~/resource", root)


def test_parse_resource_for_sharing_with_tilde_relative(root: Root) -> None:
    with pytest.raises(ValueError, match=r"Cannot expand user for "):
        parse_resource_for_sharing(f"storage:~/resource", root)


def test_parse_permission_action_read_lowercase() -> None:
    action = "read"
    assert parse_permission_action(action) == Action.READ


def test_parse_permission_action_read() -> None:
    action = "READ"
    assert parse_permission_action(action) == Action.READ


def test_parse_permission_action_write_lowercase() -> None:
    action = "write"
    assert parse_permission_action(action) == Action.WRITE


def test_parse_permission_action_write() -> None:
    action = "WRITE"
    assert parse_permission_action(action) == Action.WRITE


def test_parse_permission_action_manage_lowercase() -> None:
    action = "manage"
    assert parse_permission_action(action) == Action.MANAGE


def test_parse_permission_action_manage() -> None:
    action = "MANAGE"
    assert parse_permission_action(action) == Action.MANAGE


def test_parse_permission_action_wrong_string() -> None:
    action = "tosh"
    err = "invalid permission action 'tosh', allowed values: read, write, manage"
    with pytest.raises(ValueError, match=err):
        parse_permission_action(action)


def test_parse_permission_action_wrong_empty() -> None:
    action = ""
    err = "invalid permission action '', allowed values: read, write, manage"
    with pytest.raises(ValueError, match=err):
        parse_permission_action(action)


def test_pager_maybe_no_tty() -> None:
    with mock.patch.multiple(
        "click", echo=mock.DEFAULT, echo_via_pager=mock.DEFAULT
    ) as mocked:
        mock_echo = mocked["echo"]
        mock_echo_via_pager = mocked["echo_via_pager"]

        terminal_size = (100, 10)
        tty = False
        large_input = [f"line {x}" for x in range(20)]

        pager_maybe(large_input, tty, terminal_size)
        assert mock_echo.call_args_list == [mock.call(x) for x in large_input]
        mock_echo_via_pager.assert_not_called()


def test_pager_maybe_terminal_larger() -> None:
    with mock.patch.multiple(
        "click", echo=mock.DEFAULT, echo_via_pager=mock.DEFAULT
    ) as mocked:
        mock_echo = mocked["echo"]
        mock_echo_via_pager = mocked["echo_via_pager"]

        terminal_size = (100, 10)
        tty = True
        small_input = ["line 1", "line 2"]

        pager_maybe(small_input, tty, terminal_size)
        assert mock_echo.call_args_list == [mock.call(x) for x in small_input]
        mock_echo_via_pager.assert_not_called()


def test_pager_maybe_terminal_smaller() -> None:
    with mock.patch.multiple(
        "click", echo=mock.DEFAULT, echo_via_pager=mock.DEFAULT
    ) as mocked:
        mock_echo = mocked["echo"]
        mock_echo_via_pager = mocked["echo_via_pager"]

        terminal_size = (100, 10)
        tty = True
        large_input = [f"line {x}" for x in range(20)]

        pager_maybe(large_input, tty, terminal_size)
        mock_echo.assert_not_called()
        mock_echo_via_pager.assert_called_once()
        lines_it = mock_echo_via_pager.call_args[0][0]
        assert "".join(lines_it) == "\n".join(large_input)

        # Do the same, but call with a generator function for input instead
        mock_echo_via_pager.reset_mock()
        iter_input = iter(large_input)
        next(iter_input)  # Skip first line

        pager_maybe(iter_input, tty, terminal_size)
        mock_echo.assert_not_called()
        mock_echo_via_pager.assert_called_once()
        lines_it = mock_echo_via_pager.call_args[0][0]
        assert "".join(lines_it) == "\n".join(large_input[1:])


async def test_calc_life_span_none_default(
    monkeypatch: Any, tmp_path: Path, make_client: _MakeClient
) -> None:
    async with make_client("https://example.com") as client:
        monkeypatch.chdir(tmp_path)
        local_conf = tmp_path / ".neuro.toml"
        local_conf.write_text(toml.dumps({"job": {"life-span": "1d2h3m4s"}}))
        expected = timedelta(days=1, hours=2, minutes=3, seconds=4)
        assert (
            await calc_life_span(client, None, "1d", "job") == expected.total_seconds()
        )


async def test_calc_life_span_default_life_span_all_keys(
    caplog: Any, monkeypatch: Any, tmp_path: Path, make_client: _MakeClient
) -> None:
    async with make_client("https://example.com") as client:
        monkeypatch.chdir(tmp_path)
        local_conf = tmp_path / ".neuro.toml"
        # empty config
        local_conf.write_text(toml.dumps({"job": {"life-span": "1d2h3m4s"}}))

        expected = timedelta(days=1, hours=2, minutes=3, seconds=4)
        assert (
            await calc_life_span(client, None, "1d", "job") == expected.total_seconds()
        )


async def test_calc_default_life_span_invalid(
    caplog: Any,
    monkeypatch: Any,
    tmp_path: Path,
    make_client: _MakeClient,
) -> None:
    async with make_client("https://example.com") as client:
        monkeypatch.chdir(tmp_path)
        local_conf = tmp_path / ".neuro.toml"
        # empty config
        local_conf.write_text(toml.dumps({"job": {"life-span": "invalid"}}))
        with pytest.raises(
            click.UsageError,
            match="Could not parse time delta",
        ):
            await calc_life_span(client, None, "1d", "job")


async def test_calc_default_life_span_default_value(
    caplog: Any,
    monkeypatch: Any,
    tmp_path: Path,
    make_client: _MakeClient,
) -> None:
    async with make_client("https://example.com") as client:
        monkeypatch.chdir(tmp_path)
        local_conf = tmp_path / ".neuro.toml"
        # empty config
        local_conf.write_text(toml.dumps({}))
        default = parse_timedelta("1d")
        assert (
            await calc_life_span(client, None, "1d", "job") == default.total_seconds()
        )



import contextlib
import logging
import sqlite3
import time
from typing import Any, Dict, List, Optional, Tuple

import aiohttp
import certifi
import click
import dateutil.parser
import pkg_resources
from typing_extensions import TypedDict
from yarl import URL

import neuromation
from neuromation.api import Client


class Record(TypedDict):
    package: str
    version: str
    uploaded: float
    checked: float


log = logging.getLogger(__name__)


SCHEMA = {
    "pypi": "CREATE TABLE pypi "
    "(package TEXT, version TEXT, uploaded REAL, checked REAL)",
}
DROP = {"pypi": "DROP TABLE IF EXISTS pypi"}


async def run_version_checker(client: Client, disable_check: bool) -> None:
    if disable_check:
        return
    with client.config._open_db() as db:
        _ensure_schema(db)
        neurocli_db = _read_package(db, "neuro-cli")
        neuromation_db = _read_package(db, "neuromation")
        certifi_db = _read_package(db, "certifi")

    _warn_maybe(neurocli_db, neuromation_db, certifi_db)
    inserts: List[Tuple[str, str, float, float]] = []
    await _add_record(client, "neuromation", neuromation_db, inserts)
    await _add_record(client, "neuro-cli", neuromation_db, inserts)
    await _add_record(client, "certifi", certifi_db, inserts)
    with client.config._open_db() as db:
        db.executemany(
            """
            INSERT INTO pypi (package, version, uploaded, checked)
            VALUES (?, ?, ?, ?)
        """,
            inserts,
        )
        db.execute("DELETE FROM pypi WHERE checked < ?", (time.time() - 7 * 24 * 3600,))
        with contextlib.suppress(sqlite3.OperationalError):
            db.commit()


async def _add_record(
    client: Client,
    package: str,
    record: Optional[Record],
    inserts: List[Tuple[str, str, float, float]],
) -> None:
    if record is None or time.time() - record["checked"] > 10 * 60:
        pypi = await _fetch_package(client._session, package)
        if pypi is None:
            return
        inserts.append(
            (
                pypi["package"],
                pypi["version"],
                pypi["uploaded"],
                pypi["checked"],
            )
        )


def _ensure_schema(db: sqlite3.Connection) -> None:
    cur = db.cursor()
    ok = True
    found = set()
    cur.execute("SELECT type, name, sql from sqlite_master")
    for type, name, sql in cur:
        if type not in ("table", "index"):
            continue
        if name in SCHEMA:
            if SCHEMA[name] != sql:
                ok = False
                break
            else:
                found.add(name)

    if not ok or found < SCHEMA.keys():
        for sql in reversed(list(DROP.values())):
            cur.execute(sql)
        for sql in SCHEMA.values():
            cur.execute(sql)


READ_PACKAGE = """
    SELECT package, version, uploaded, checked
    FROM pypi
    WHERE package = ?
    ORDER BY checked
    LIMIT 1
"""


def _read_package(db: sqlite3.Connection, package: str) -> Optional[Record]:
    cur = db.execute(READ_PACKAGE, (package,))
    return cur.fetchone()


async def _fetch_package(
    session: aiohttp.ClientSession, package: str
) -> Optional[Record]:
    url = URL(f"https://pypi.org/pypi/{package}/json")
    async with session.get(url) as resp:
        if resp.status != 200:
            log.debug("%s status on fetching %s", resp.status, url)
            return None
        pypi_response = await resp.json()
        version = _parse_max_version(pypi_response)
        if version is None:
            return None
        uploaded = _parse_version_upload_time(pypi_response, version)
        return {
            "package": package,
            "version": version,
            "uploaded": uploaded,
            "checked": time.time(),
        }


def _parse_date(value: str) -> float:
    # from format: "2019-08-19"
    return dateutil.parser.parse(value).timestamp()


def _parse_max_version(pypi_response: Dict[str, Any]) -> Optional[str]:
    try:
        ret = [version for version in pypi_response["releases"].keys()]
        return max(
            ver for ver in ret if not pkg_resources.parse_version(ver).is_prerelease  # type: ignore  # noqa
        )
    except (KeyError, ValueError):
        return None


def _parse_version_upload_time(
    pypi_response: Dict[str, Any], target_version: str
) -> float:
    try:
        dates = [
            _parse_date(info["upload_time"])
            for version, info_list in pypi_response["releases"].items()
            for info in info_list
            if version == target_version
        ]
        return max(dates)
    except (KeyError, ValueError):
        return 0


def _warn_maybe(
    neurocli_db: Optional[Record],
    neuromation_db: Optional[Record],
    certifi_db: Optional[Record],
    *,
    certifi_warning_delay: int = 14 * 3600 * 24,
) -> None:

    if neurocli_db is not None:
        current = pkg_resources.parse_version(neuromation.__version__)
        pypi = pkg_resources.parse_version(neurocli_db["version"])
        if current < pypi:
            update_command = "pip install --upgrade neuro-cli"
            click.secho(
                f"You are using Neuro Platform Client {current}, "
                f"however {pypi} is available.\n"
                f"You should consider upgrading via "
                f"the '{update_command}' command.",
                err=True,
                fg="yellow",
            )
    elif neuromation_db is not None:
        current = pkg_resources.parse_version(neuromation.__version__)
        pypi = pkg_resources.parse_version(neuromation_db["version"])
        if current < pypi:
            update_command = "pip install --upgrade neuromation"
            click.secho(
                f"You are using Neuro Platform Client {current}, "
                f"however {pypi} is available.\n"
                f"You should consider upgrading via "
                f"the '{update_command}' command.",
                err=True,
                fg="yellow",
            )

    if certifi_db is not None:
        current = pkg_resources.parse_version(certifi.__version__)  # type: ignore
        pypi = pkg_resources.parse_version(certifi_db["version"])
        if (
            current < pypi
            and time.time() - certifi_db["uploaded"] > certifi_warning_delay
        ):
            pip_update_command = "pip install --upgrade certifi"
            conda_update_command = "conda update certifi"
            click.secho(
                f"Your root certificates are out of date.\n"
                f"You are using certifi {current}, "
                f"however {pypi} is available.\n"
                f"Please consider upgrading certifi package, e.g.\n"
                f"    {pip_update_command}\n"
                f"or\n"
                f"    {conda_update_command}",
                err=True,
                fg="red",
            )



from prompt_toolkit.key_binding import KeyPress
from prompt_toolkit.keys import Keys

from neuromation.cli.ael import _has_detach


def test_detach_short() -> None:
    term = (Keys.ControlP, Keys.ControlQ)
    assert not _has_detach([], term)


def test_detach_not_present() -> None:
    term = (Keys.ControlP, Keys.ControlQ)
    assert not _has_detach(
        [
            KeyPress(Keys.Control0, ""),
            KeyPress(Keys.ControlP, ""),
            KeyPress(Keys.Control1, ""),
        ],
        term,
    )


def test_detach_with_other_symbols() -> None:
    term = (Keys.ControlP, Keys.ControlQ)
    assert not _has_detach(
        [
            KeyPress(Keys.Control0, ""),
            KeyPress(Keys.ControlP, ""),
            KeyPress(Keys.Control1, ""),
            KeyPress(Keys.ControlQ, ""),
        ],
        term,
    )


def test_detach_reversed() -> None:
    term = (Keys.ControlP, Keys.ControlQ)
    assert not _has_detach(
        [
            KeyPress(Keys.Control0, ""),
            KeyPress(Keys.ControlQ, ""),
            KeyPress(Keys.ControlP, ""),
            KeyPress(Keys.Control1, ""),
        ],
        term,
    )


def test_detach_present_at_the_end() -> None:
    term = (Keys.ControlP, Keys.ControlQ)
    assert _has_detach(
        [
            KeyPress(Keys.Control0, ""),
            KeyPress(Keys.ControlP, ""),
            KeyPress(Keys.ControlQ, ""),
        ],
        term,
    )


def test_detach_present_in_the_middle() -> None:
    term = (Keys.ControlP, Keys.ControlQ)
    assert _has_detach(
        [
            KeyPress(Keys.Control0, ""),
            KeyPress(Keys.ControlP, ""),
            KeyPress(Keys.ControlQ, ""),
            KeyPress(Keys.Control1, ""),
        ],
        term,
    )



from typing import Callable

from yarl import URL

from neuromation.api import RemoteImage
from neuromation.api.url_utils import CLUSTER_SCHEMES


URIFormatter = Callable[[URL], str]
ImageFormatter = Callable[[RemoteImage], str]


def uri_formatter(username: str, cluster_name: str) -> URIFormatter:
    def formatter(uri: URL) -> str:
        if uri.scheme in CLUSTER_SCHEMES:
            if uri.host == cluster_name:
                assert uri.path[0] == "/"
                path = uri.path.lstrip("/")
                owner, _, rest = path.partition("/")
                if owner == username:
                    return f"{uri.scheme}:{rest.lstrip('/')}"
                return f"{uri.scheme}:/{path}"
        return str(uri)

    return formatter


def image_formatter(uri_formatter: URIFormatter) -> ImageFormatter:
    def formatter(image: RemoteImage) -> str:
        image_str = str(image)
        if image_str.startswith("image://"):
            return uri_formatter(URL(image_str))
        else:
            return image_str

    return formatter


def yes() -> str:
    return "[green]√[/green]"


def no() -> str:
    return "[red]×[/red]"
